"""Run all problems and associated examples in espresso

This script assumes you have geo-espresso installed via:
$ python espresso_machine/build_package/build.py

"""

import sys
import os
import typing
import types
import pathlib
import subprocess
import dataclasses

import _utils

try:
    from espresso.exceptions import InvalidExampleError
    from espresso import EspressoProblem
except ModuleNotFoundError as e:
    e.msg += (
        "\n\nNote: To run pre-build validation, please firstly install "
        "`espresso` core module by running the following from the root"
        "level of the project\n  $ pip install ."
    )
    raise e


@dataclasses.dataclass
class ResultsFromExample:
    prob_instance_str: str  # e.g. "SimpleRegression(2)""
    prob_instance: EspressoProblem  # e.g. actual instance of str above
    i: int  # e.g. 2
    results: dict

    def error_in_init(
        self,
    ) -> typing.Tuple[str, typing.Union[Exception, EspressoProblem]]:
        return isinstance(self.prob_instance, Exception), self.prob_instance

    def __getattr__(self, key):
        if key in self.results:
            return self.results[key]
        else:
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute '{key}'"
            )


@dataclasses.dataclass
class ResultsFromProblem:
    parent_module: types.ModuleType
    problem_name: str
    problem_class: typing.Type
    problem_class_str: str
    problem_path: str
    problem_results_generator: typing.Iterator[ResultsFromExample]


class _ProblemModule:
    """get the parent module of a problem class
    two different cases:
    1. run the examples from contrib/ (pre-build)
    2. run the examples from a built espresso package (post-build)
    """

    def __init__(self, pre_build, problem_name):
        self._pre_build = pre_build
        self._problem_name = problem_name

    def __enter__(self):
        if self._pre_build:
            sys.path.insert(1, _utils.CONTRIB_FOLDER)
            return __import__(self._problem_name)
        else:
            return __import__(_utils.PKG_NAME)

    def __exit__(self, exc_type, exc_value, traceback):
        _to_del = set()
        for key in sys.modules.keys():
            if self._problem_name in key:
                _to_del.add(key)
        for m in _to_del:
            del sys.modules[m]


# For each of the below methods / properties,
# 1. Try to get / access
# 2. If not implemented, assign None to `output_name_for_testing_purpose`
# 3. If implemeneted but got error, assign the error to `output_name_for_testing_purpose`
# 4. If implemeneted and no error, assign the output to `output_name_for_testing_purpose`

prob_methods = [
    # (output_name_for_testing_purpose, how_to_get_it)
    ("synth1", lambda p: p.forward(p.good_model)),
    ("jac1", lambda p: p.jacobian(p.good_model)),
    ("synth2", lambda p: p.forward(p.good_model, True)[0]),
    ("jac2", lambda p: p.forward(p.good_model, True)[1]),
    ("fig_model", lambda p: p.plot_model(p.good_model)),
    ("fig_data", lambda p: p.plot_data(p.data)),
    ("misfit", lambda p: p.misfit(p.data, p.data)),
    ("log_likelihood", lambda p: p.log_likelihood(p.data, p.data)),
    ("log_prior", lambda p: p.log_prior(p.good_model)),
]
prob_properties = [
    # (output_name_for_testing_purpose, how_to_access_it)
    ("nmodel", "model_size"),
    ("ndata", "data_size"),
    ("model", "good_model"),
    ("null_model", "starting_model"),
    ("data", "data"),
    ("description", "description"),
    ("cov", "covariance_matrix"),
    ("inv_cov", "inverse_covariance_matrix"),
]


def _get_result(prob_instance, how_to_get, is_method=True, timeout=None):
    @_utils.timeout(seconds=timeout)
    def _run_attr(prob_instance, how_to_get, is_method=True):
        if is_method:
            return how_to_get(prob_instance)
        return getattr(prob_instance, how_to_get)

    try:
        return _run_attr(prob_instance, how_to_get, is_method)
    except NotImplementedError:
        return None
    except Exception as e:
        return e


def instantiate_example(problem_class, i):
    try:
        prob_instance_i = problem_class(i)
    except Exception as e:
        if not isinstance(e, InvalidExampleError):
            prob_instance_i = e
        else:
            raise e
    return prob_instance_i


def collect_methods_outputs(prob_instance_i, all_outputs, timeout=None):
    for (output_name, how) in prob_methods:
        all_outputs[output_name] = _get_result(prob_instance_i, how, True, timeout)


def collect_properties(prob_instance_i, all_outputs, timeout=None):
    for (output_name, prop) in prob_properties:
        all_outputs[output_name] = _get_result(prob_instance_i, prop, False, timeout)


def run_example(
    problem_class, problem_class_str, i, timeout=None
) -> ResultsFromExample:
    # prepare
    all_outputs = dict()
    prob_instance_i = instantiate_example(problem_class, i)
    results_from_example = ResultsFromExample(
        prob_instance_str=f"{problem_class_str}({i})",
        prob_instance=prob_instance_i,
        i=i,
        results=all_outputs,
    )
    if not results_from_example.error_in_init()[0]:  # collect results
        collect_methods_outputs(prob_instance_i, all_outputs, timeout)
        collect_properties(prob_instance_i, all_outputs, timeout)
    return results_from_example


def run_problem(
    problem_class, problem_class_str, timeout=None
) -> typing.Iterator[dict]:
    if isinstance(problem_class, Exception):
        return []
    i = 1
    while True:
        if i > 99:
            raise ValueError(
                "Reached example 100: aborting."
            )  # Guard against silliness
        try:
            example_res = run_example(problem_class, problem_class_str, i, timeout)
        except InvalidExampleError:
            if i == 1:
                raise ValueError("Ensure there are at least one examples")
            return
        if example_res.error_in_init()[0]:
            raise example_res.error_in_init()[1]
        i += 1
        yield example_res


def run_cmake_if_needed(prob_path, pre_build):
    if pre_build and "CMakeLists.txt" in os.listdir(prob_path):
        build_dir = pathlib.Path(prob_path) / "build"
        build_dir.mkdir(exist_ok=True)
        res1 = subprocess.call(["cmake", ".."], cwd=build_dir)
        if res1:
            raise ChildProcessError(f"`cmake ..` failed in {prob_path}/build")
        res2 = subprocess.call(["make"], cwd=build_dir)
        if res2:
            raise ChildProcessError(f"`make` failed in {prob_path}/build")


def run_problems(
    problems, pre_build, timeout=None
) -> typing.Iterator[ResultsFromProblem]:
    for (prob_name, prob_path) in problems:
        prob_class_str = _utils.problem_name_to_class(prob_name)
        run_cmake_if_needed(prob_path, pre_build)
        try:
            with _ProblemModule(pre_build, prob_name) as parent_module:
                try:
                    prob_class = getattr(parent_module, prob_class_str)
                except Exception as e:
                    prob_class = e
                yield ResultsFromProblem(
                    parent_module=parent_module,
                    problem_name=prob_name,
                    problem_class=prob_class,
                    problem_class_str=prob_class_str,
                    problem_path=prob_path,
                    problem_results_generator=run_problem(
                        prob_class, prob_class_str, timeout
                    ),
                )
        except Exception as e:
            yield ResultsFromProblem(
                parent_module=e,
                problem_name=prob_name,
                problem_class=None,
                problem_class_str=prob_class_str,
                problem_path=prob_path,
                problem_results_generator=[],
            )


def main(problems_specified=None, timeout=_utils.DEFAULT_TIMEOUT, verbose=False):
    problems = _utils.problems_to_run(problems_specified)
    results = run_problems(problems, pre_build=True, timeout=timeout)
    for res in results:
        if verbose:
            print(res.problem_class)
        for prob_out_i in res.problem_results_generator:
            if verbose:
                print(prob_out_i.keys())


if __name__ == "__main__":
    main()



"""Generate API compliance report 

Usage:
- To generate raw report, raw_compliance_report(problems_to_check=None, pre_build=True)
- To generate report, compliance_report(problems_to_check=None, pre_build=True)
- To print report, print_compliance_report(report)
"""
import dataclasses
import typing

import run_examples
import criteria
import _utils


@dataclasses.dataclass
class ProblemRawReport:
    problem_name: str
    problem_class_name: str
    metadata: typing.Union[bool, Exception]
    attributes_required: dict[str, list[typing.Optional[bool]]]
    attributes_optional: dict[str, list[typing.Optional[bool]]]
    attributes_additional: list[str]

    def __init__(self, results: run_examples.ResultsFromProblem):
        self.problem_name = results.problem_name
        self.problem_class_name = results.problem_class_str
        self.metadata = self._collect_metadata_report(results)
        if self.metadata_ok():
            self.attributes_required = dict()
            self.attributes_optional = dict()
            self.attributes_additional = list()
            self._collect_attributes_report(results)

    def metadata_ok(self) -> bool:
        return self.metadata == True

    def _collect_metadata_report(self, results: run_examples.ResultsFromProblem):
        if isinstance(results.parent_module, Exception):
            return results.parent_module
        else:
            try:
                criteria.criteria_for_problem(
                    results.problem_class,
                    results.problem_class_str,
                    results.problem_path,
                    results.parent_module,
                )
            except Exception as e:
                return e
            else:
                return True

    def _collect_attributes_report(self, results: run_examples.ResultsFromProblem):
        for results_i in results.problem_results_generator:
            self._collect_compliance_info(results_i)
            self._collect_additional_info(results_i)

    def _collect_compliance_info(self, results_i: run_examples.ResultsFromExample):
        _has_init_error = isinstance(results_i.prob_instance, Exception)
        if _has_init_error:
            _init_error = results_i.prob_instance
        for attr_check in criteria.attributes_to_check:
            attr_key, attr_str, required, to_check = attr_check
            _report_key = "attributes_required" if required else "attributes_optional"
            _attr_name = attr_str.split("(")[0]
            if _attr_name not in getattr(self, _report_key):
                getattr(self, _report_key)[_attr_name] = []
            _to_update = getattr(self, _report_key)[_attr_name]
            if _has_init_error:
                _to_update.append(_init_error)
                continue
            obj = getattr(results_i, attr_key)
            obj_str = f"{results_i.prob_instance_str}.{attr_str}"
            if isinstance(obj, Exception) or obj is None:
                _to_update.append(obj)
            else:
                try:
                    for check_func in to_check:
                        check_func(results_i, obj, obj_str)
                except Exception as e:
                    _to_update.append(e)
                else:
                    _to_update.append(True)

    def _get_standard_attr(self, problem_instance):
        standard_attr = set(self.attributes_required.keys())
        standard_attr.update(self.attributes_optional.keys())
        standard_attr.update(problem_instance.__abstract_metadata_keys__)
        standard_attr.update({"metadata", "params", "example_number"})
        return standard_attr

    def _collect_additional_info(self, results_i: run_examples.ResultsFromExample):
        p = results_i.prob_instance
        if isinstance(p, Exception):
            return
        dir_p = [attr for attr in dir(p) if not attr.startswith("_")]
        _standard_attr = self._get_standard_attr(p)
        _additional_attr = {a for a in dir_p if a not in _standard_attr}
        self.attributes_additional = _additional_attr


def raw_compliance_report(problems_to_check=None, pre_build=True, timeout=None) \
    -> dict[str, ProblemRawReport]:
    """Run all problems and generate a raw compliance report

    A typical raw report looks like:
    {
        'FmmTomography': {
            'metadata': True,
            'attributes_required': {
                'model_size': [True], 'data_size': [True], 'good_model': [True], 'starting_model': [True], 'data': [True], 'forward': [True]
            },
            'attributes_optional': {
                'description': [None], 'covariance_matrix': [None], 'inverse_covariance_matrix': [None], 'jacobian': [True], 'forward': [True, True], 'plot_model': [True], 'plot_data': [None], 'misfit': [None], 'log_likelihood': [None], 'log_prior': [None]
            },
            'attributes_additional': {
                'exe_fm2dss', 'clean_tmp_files', 'tmp_paths', 'tmp_files'
            }
        }
    }
    """
    raw_report = dict()
    problems = _utils.problems_to_run(problems_to_check)
    results = run_examples.run_problems(problems, pre_build=pre_build, timeout=timeout)
    for res in results:
        raw_report[res.problem_class_str] = ProblemRawReport(res)
    return raw_report


@dataclasses.dataclass
class ProblemReport:
    problem_name: str
    problem_class_name: str
    metadata: typing.Union[str, Exception]
    required: dict[str, typing.Union[str, Exception]]
    required_count: dict[str, int]
    optional: dict[str, typing.Union[str, Exception]]
    optional_count: dict[str, int]
    additional: list[str]
    additional_count: int
    api_compliance: bool

    def __init__(self, raw_report: ProblemRawReport):
        self.problem_name = raw_report.problem_name
        self.problem_class_name = raw_report.problem_class_name
        self.metadata = self._collect_metadata_report(raw_report)
        if self.metadata_ok():
            self._collect_required_attr_report(raw_report)
            self._collect_optional_attr_report(raw_report)
            self._collect_additional_attr_report(raw_report)
            self._analyse_compliance()
        else:
            self.api_compliance = self.metadata

    def metadata_ok(self) -> bool:
        return self.metadata == "OK"
    
    def _collect_metadata_report(self, raw_report: ProblemRawReport):
        return "OK" if raw_report.metadata_ok() else raw_report.metadata

    @staticmethod
    def _analyse_report_dict(sub_report):
        _new_report = dict()
        _new_count = {
            "implemented": 0,
            "not_implemented": 0,
            "error": 0,
            "total": 0,
        }
        for item_name, item_report in sub_report.items():
            _done = [r for r in item_report if r == True]
            _has_error = [r for r in item_report if isinstance(r, Exception)]
            if len(_done) == len(item_report):
                _new_count["implemented"] += 1
                _new_report[item_name] = "OK"
            elif len(_has_error) > 0:
                _new_count["error"] += 1
                _new_report[item_name] = _has_error[0]
            else:
                _new_count["not_implemented"] += 1
                _new_report[item_name] = "Not implemented"
            _new_count["total"] += 1
        return {"report": _new_report, "count": _new_count}

    def _collect_required_attr_report(self, raw_report: ProblemRawReport):
        res = self._analyse_report_dict(raw_report.attributes_required)
        self.required = res["report"]
        self.required_count = res["count"]

    def _collect_optional_attr_report(self, raw_report: ProblemRawReport):
        res = self._analyse_report_dict(raw_report.attributes_optional)
        self.optional = res["report"]
        self.optional_count = res["count"]
    
    def _collect_additional_attr_report(self, raw_report: ProblemRawReport):
        self.additional = raw_report.attributes_additional
        self.additional_count = len(raw_report.attributes_additional)
    
    def _analyse_compliance(self):
        _metadata_ok = self.metadata_ok()
        _required_ok = \
            self.required_count["implemented"] == self.required_count["total"]
        _optional_ok = self.optional_count["error"] == 0
        self.api_compliance = _metadata_ok and _required_ok and _optional_ok


def compliance_report(problems_to_check=None, pre_build=True, timeout=_utils.DEFAULT_TIMEOUT) \
    -> dict[str, ProblemReport]:
    """Generate a readable compliance report based on running raw report

    A typical compliance report looks like:
    {
        'FmmTomography': {
            'metadata': 'OK',
            'required': {'model_size': 'OK', 'data_size': 'OK', 'good_model': 'OK', 'starting_model': 'OK', 'data': 'OK', 'forward': 'OK'},
            'required_count': {'implemented': 6, 'not_implemented': 0, 'error': 0, 'total': 6},
            'optional': {'description': 'Not implemented', 'covariance_matrix': 'Not implemented', 'inverse_covariance_matrix': 'Not implemented', 'jacobian': 'OK', 'forward': 'OK', 'plot_model': 'OK', 'plot_data': 'Not implemented', 'misfit': 'Not implemented', 'log_likelihood': 'Not implemented', 'log_prior': 'Not implemented'},
            'optional_count': {'implemented': 3, 'not_implemented': 7, 'error': 0, 'total': 10},
            'additional': {'clean_tmp_files', 'tmp_paths', 'exe_fm2dss', 'tmp_files'},
            'additional_count': 4,
            'api_compliance': True,
        }
    }
    """
    raw_report = raw_compliance_report(problems_to_check, pre_build, timeout)
    report = dict()
    for prob_name, prob_raw_report in raw_report.items():
        report[prob_name] = ProblemReport(prob_raw_report)
    return report


# from SO: https://stackoverflow.com/a/287944
class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def cformat(style, content):
    return f"{style}{content}{bcolors.ENDC}"


def pprint_compliance_report(report: dict[str, ProblemReport]):
    """Pretty print a compliance report (i.e. output of compliance_report())

    The console output typically looks like:
    ```
    ----------------------------------------
    ESPRESSO Machine - API Compliance Report
    ----------------------------------------

    FmmTomography
    Metadata: OK
    Required attributes: 6/6 implemented, 0 errors, 0 not implemented yet
            .model_size     - OK
            .data_size      - OK
            .good_model     - OK
            .starting_model - OK
            .data   - OK
            .forward        - OK
    Optional attributes: 3/10 implemented, 0 errors, 7 not implemented yet
            .description    - Not implemented
            .covariance_matrix      - Not implemented
            .inverse_covariance_matrix      - Not implemented
            .jacobian       - OK
            .forward        - OK
            .plot_model     - OK
            .plot_data      - Not implemented
            .misfit - Not implemented
            .log_likelihood - Not implemented
            .log_prior      - Not implemented
    Additional attributes: 4 detected
            .clean_tmp_files
            .tmp_paths
            .exe_fm2dss
            .tmp_files

    FmmTomography (fmm_tomography) is API-compliance. Cheers!
    ```
    """
    # title
    _title = "ESPRESSO Machine - API Compliance Report"
    print("-" * len(_title))
    print(cformat(bcolors.HEADER, _title))
    print("-" * len(_title))
    for prob, r in report.items():
        print()
        #### problem name
        print(cformat(bcolors.OKCYAN, cformat(bcolors.BOLD, prob)))
        #### metadata
        _metadata = cformat(bcolors.UNDERLINE, "Metadata") + ": "
        if r.metadata_ok():
            _metadata += cformat(bcolors.OKGREEN, "OK")
        else:
            _metadata += cformat(bcolors.FAIL, r.metadata)
        print(_metadata)
        #### check possibility to continue
        if not r.metadata_ok():
            continue
        #### required
        _required_title = cformat(bcolors.UNDERLINE, "Required attributes") + ": "
        _required_count = r.required_count
        _required_title += (
            f"{_required_count['implemented']}/{_required_count['implemented']} "
            f"implemented, {_required_count['error']} errors, "
            f"{_required_count['not_implemented']} not implemented yet"
        )
        print(_required_title)
        for attr_name, attr_res in r.required.items():
            if attr_res == "OK":
                attr_res_str = cformat(bcolors.OKGREEN, "OK")
            else:
                attr_res_str = cformat(bcolors.FAIL, attr_res.__repr__())
            print(f"\t.{attr_name}\t- {attr_res_str}")
        #### optional
        _optional_title = cformat(bcolors.UNDERLINE, "Optional attributes") + ": "
        _optional_count = r.optional_count
        _optional_title += (
            f"{_optional_count['implemented']}/{_optional_count['total']} "
            f"implemented, {_optional_count['error']} errors, "
            f"{_optional_count['not_implemented']} not implemented yet"
        )
        print(_optional_title)
        for attr_name, attr_res in r.optional.items():
            if attr_res == "OK":
                attr_res_str = cformat(bcolors.OKGREEN, "OK")
            elif isinstance(attr_res, Exception):
                attr_res_str = cformat(bcolors.FAIL, attr_res.__repr__())
            else:
                attr_res_str = cformat(bcolors.WARNING, attr_res)
            print(f"\t.{attr_name}\t- {attr_res_str}")
        #### additional
        _additional_title = cformat(bcolors.UNDERLINE, "Additional attributes") + ": "
        _additional_title += f"{r.additional_count} detected"
        print(_additional_title)
        for attr_name in r.additional:
            print(f"\t.{attr_name}")
        #### sum up
        _api_compliance = r.api_compliance
        if _api_compliance:
            print(cformat(bcolors.OKCYAN, f"\n{prob} ({r.problem_name}) is API-compliant. Cheers!"))
        else:
            print(cformat(bcolors.FAIL, f"\n{prob} ({r.problem_name}) is not API-compliant."))


def capability_report(problems_to_check=None, timeout=_utils.DEFAULT_TIMEOUT_SHORT):
    # those with TimeoutError will be approximated to be OK
    _compliance_report = compliance_report(problems_to_check, True, timeout)
    _capability_report = dict()
    for prob, res in _compliance_report.items():
        _new_report = dict()
        if not res.metadata_ok():
            raise RuntimeError(res)
        for attr, attr_res in res.required.items():
            _new_report[attr] = int(
                attr_res == "OK" or isinstance(attr_res, TimeoutError)
            )
        for attr, attr_res in res.optional.items():
            _new_report[attr] = int(
                attr_res == "OK" or isinstance(attr_res, TimeoutError)
            )
        for additional in res.additional:
            _new_report[additional] = 1
        _capability_report[prob] = _new_report
    return _capability_report



import numpy as np
from espresso import EspressoProblem
from espresso.exceptions import InvalidExampleError


class SimpleRegression(EspressoProblem):
    """Forward simulation class
    """

    metadata = {
        "problem_title": "Simple 1D regression",
        "problem_short_description": "This example includes various small 1D regression (curve-fitting) " \
                                "problems. Several different forms of basis function are supported: " \
                                "polynomial, Fourier and discrete.",

        "author_names": ["Andrew Valentine"],

        "contact_name": "Andrew Valentine",
        "contact_email": "andrew.valentine@durham.ac.uk",

        "citations": [],

        "linked_sites": [],
    }
    

    def __init__(self, example_number=1):
        super().__init__(example_number)
        if self.example_number == 1:
            self._xp = xp1.copy()
            self._yp = yp1.copy()
            self._m = np.array([0.5,2])
            self._basis = 'polynomial'
            self._desc = "Fitting a straight line to data"
            self._sigma = 0.1
        elif self.example_number == 2:
            self._xp = xp2.copy()
            self._yp = yp2.copy()
            self._m = np.array([1.,-0.2,0.5,0.3])
            self._basis = 'polynomial'
            self._desc = "Fitting a cubic polynomial to data"
            self._sigma = 0.1
        elif self.example_number == 3:
            self._xp = xp3.copy()
            self._yp = yp3.copy()
            self._m = np.array([1,0,0.3,-0.2,0,0,0.7,0.,0.,0.3,0.,0.,0.,0.,-.2])
            self._basis = 'fourier'
            self._desc = "Fitting a Fourier series to data"
            self._sigma = 0.1
        elif self.example_number == 4:
            self._xp = xp4.copy()
            self._yp = yp4.copy()
            self._m = np.array([0.,2,-.3,-.6])
            self._basis = 'polynomial'
            self._desc = "Fitting a polynomial to a very small dataset"
            self._sigma = 0.05
        elif self.example_number == 5:
            self._xp = xp5.copy()
            self._yp = yp5.copy()
            self._m = np.array([0.3,0.3,0.,-0.2,0.5,-.8,0.1,0.125])
            self._basis = 'polynomial'
            self._desc = "Fitting a polynomial to a dataset with a gap"
            self._sigma = 0.1
        else:
            raise InvalidExampleError

    @property
    def description(self):
        return self._desc

    @property
    def model_size(self):
        return self._m.size

    @property
    def data_size(self):
        return self._yp.size

    @property
    def good_model(self):
        return self._m.copy()

    @property
    def starting_model(self):
        return np.zeros_like(self._m)
    
    @property
    def data(self):
        return self._yp.copy()

    @property
    def covariance_matrix(self):
        return self._sigma**2 * np.eye(self.data_size)

    @property
    def inverse_covariance_matrix(self):
        return (1/self._sigma**2) * np.eye(self.data_size)
        
    def forward(self, model, with_jacobian=False):
        d = curveFittingFwd(self._m,self._xp,self._basis)
        if with_jacobian:
            return d, self.jacobian(model)
        else:
            return d
    
    def jacobian(self, model):
        nModel = model.shape[0]
        return curveFittingJac(self._xp,nModel,self._basis)

    def plot_model(self, model):
        raise NotImplementedError               # optional
    
    def plot_data(self, data, data2=None):
        raise NotImplementedError               # optional

    def misfit(self, data, data2):              # optional
        raise NotImplementedError

    def log_likelihood(self,data,data2):        # optional
        raise NotImplementedError
    
    def log_prior(self, model):                 # optional
        raise NotImplementedError

# The following arrays define the 'sampling points' used in various examples
# In principle we could generate these on-the-fly but using hard-coded values
# guarantees repeatibility across different machines, architectures, prngs.

# The first four are uniformly-distributed points between 0 and 1
xp1 = np.array([0.02814768, 0.0885654 , 0.11748064, 0.12294414, 0.16288759,
                0.16877703, 0.1843951 , 0.39127594, 0.46038946, 0.4914394 ,
                0.54084596, 0.5641636 , 0.58900649, 0.58976937, 0.6547135 ,
                0.85518813, 0.86779304, 0.91481368, 0.951955  , 0.98545064])
xp2 = np.array([0.02061079, 0.02454563, 0.19277473, 0.2784715 , 0.29414602,
                0.32944733, 0.42601344, 0.42966824, 0.45568706, 0.4727606 ,
                0.47873683, 0.50393437, 0.59354021, 0.60901867, 0.68636357,
                0.70977542, 0.73149133, 0.78673413, 0.82707107, 0.84599549,
                0.88076936, 0.9171245 , 0.95712395, 0.97514086, 0.99942137])
xp3 = np.array([0.02129797, 0.0680587 , 0.07325879, 0.07540233, 0.0815562 ,
                0.11653215, 0.13302591, 0.17036322, 0.17948113, 0.18668185,
                0.22055489, 0.22751703, 0.26470851, 0.31516965, 0.34038959,
                0.34051367, 0.34721832, 0.35517515, 0.36644436, 0.42221368,
                0.4780765 , 0.50384201, 0.54011153, 0.56945004, 0.65217677,
                0.65762461, 0.66908502, 0.68851014, 0.71684459, 0.717921  ,
                0.72093096, 0.73367274, 0.73389493, 0.75033087, 0.75890497,
                0.76225345, 0.76552936, 0.83833901, 0.86436217, 0.88042872,
                0.90603222, 0.91094849, 0.9336106 , 0.9400528 , 0.96037445,
                0.965113  , 0.96609565, 0.97968897, 0.98044997, 0.99266562])
xp4 = np.array([0.17833636, 0.19050344, 0.26430464, 0.51520092, 0.93174146])
# This set is made up of 25 samples uniformly-distributed in [0,0.3] and a further
# 5 samples uniformly-distributed in [0.9,1]
xp5 = np.array([0.01394879, 0.02194952, 0.02925089, 0.03395566, 0.05389559,
                0.0754086 , 0.08002495, 0.08598053, 0.09475561, 0.12191561,
                0.12315494, 0.14364847, 0.18455025, 0.19247459, 0.19436732,
                0.20205877, 0.20425212, 0.21203449, 0.21238893, 0.2316263 ,
                0.2381924 , 0.24158175, 0.260489  , 0.26776801, 0.27662425,
                0.90844754, 0.92687324, 0.95818959, 0.98446575, 0.99313721])

# The following arrays define the 'data' at these sampling points.
# yp1 = curveFittingFwd(np.array([0.5,2]),xp1,'polynomial') + np.random.normal(0,0.1,size=20)
yp1 = np.array([0.62075831, 0.66404134, 0.6541734 , 0.73098024, 0.98690644,
                0.96543423, 0.72329687, 1.39495485, 1.25161344, 1.32057621,
                1.49919959, 1.64859318, 1.67027886, 1.64236188, 1.75145179,
                2.15888014, 2.24377283, 2.29298466, 2.55081749, 2.32692856])
# yp2 = curveFittingFwd(np.array([1.,-0.2,0.5,0.3]),xp2,'polynomial') + np.random.normal(0,0.1,size=25)
yp2 = np.array([1.08325397, 1.03071536, 0.94623718, 0.91563318, 0.9506136 ,
                0.9367868 , 0.89703371, 1.02743885, 1.09065488, 1.08749366,
                1.06110559, 1.12980868, 1.14805172, 1.1982756 , 1.09985941,
                1.2375668 , 1.16813041, 1.28086391, 1.29770187, 1.51483868,
                1.40831814, 1.54380668, 1.49585496, 1.60347859, 1.58731048])
# yp3 = curveFittingFwd(np.array([1,0,0.3,-0.2,0,0,0.7,0.,0.,0.3,0.,0.,0.,0.,-.2]),xp3,'fourier') + np.random.normal(0,0.1,size=50)
yp3 = np.array([ 1.59943492,  1.29476968,  1.12470341,  1.21775137,  0.97557667,
                -0.30627497, -0.60216963, -0.44427947, -0.40124672, -0.35675986,
                0.47297917,  0.78227069,  0.97731201,  1.01290847,  1.06372535,
                0.99466542,  1.02487743,  0.9231454 ,  1.14490614,  0.42386911,
                -0.16176363, -0.33208629, -0.70905047, -0.46227597,  1.37939264,
                1.2141395 ,  1.21686748,  0.92069665,  0.54537001,  0.3790561 ,
                0.39905301,  0.20934235,  0.15370332,  0.29606956,  0.19790313,
                0.21644467,  0.21123332,  0.27204311,  0.39561473,  0.55357158,
                0.7340072 ,  0.8733571 ,  1.11949072,  1.26242578,  1.11077514,
                0.97156732,  1.05720015,  1.14813579,  1.3159233 ,  1.29343741])
# yp4 = curveFittingFwd(np.array([0.,2,-.3,-.6]),xp4,'polynomial') + np.random.normal(0,0.05,5)
yp4 = np.array([0.36882181, 0.43255407, 0.51268123, 0.94149677, 1.07863492])
#yp5 = curveFittingFwd(np.array([0.3,0.3,0.,-0.2,0.5,-.8,0.1,0.125]),xp5,'polynomial') + np.random.normal(0,0.1,30)
yp5 = np.array([0.32980738, 0.39458956, 0.24584324, 0.44555868, 0.31664395,
                0.35942105, 0.28608304, 0.16866817, 0.30498625, 0.24166827,
                0.25188134, 0.31265473, 0.33615641, 0.31067832, 0.29573255,
                0.37717869, 0.25847226, 0.54623771, 0.40138535, 0.22123748,
                0.32116732, 0.38347333, 0.43804293, 0.31016492, 0.34991239,
                0.38015441, 0.25919217, 0.25775872, 0.23213485, 0.41639483])

def curveFittingFwd(model, xpts, basis='polynomial',domainLength = 1.):
    """
    A fairly generic forward model for a range of 1D curve-fitting problems.
    model        - array of model parameters
    xpts         - array of sampling points
    basis        - one of "polynomial", "fourier" or "discrete"
    domainLength - float; all valid values for xpts should be in the range [0,domainLength]
    """
    singlePoint = False
    if domainLength<=0: raise ValueError("Argument 'domainLength' must be positive")
    # Convert list inputs to arrays
    if type(model) is type([]): model = np.array(model)
    if type(xpts) is type([]): xpts = np.array(xpts)
    # Check model parameters
    try:
        nModelParameters = model.shape[0]
    except AttributeError:
        raise ValueError("Argument 'model' should be a 1-D array")
    if len(model.shape)>1: 
        raise ValueError("Argument 'model' should be a 1-D array")
    # Check x-values
    try:
        npts = xpts.shape[0]
        if len(xpts.shape)!=1: raise ValueError("Argument 'xpts' should be a 1-D array")
    except AttributeError:
        singlePoint = True
        npts = 1
    if basis == 'polynomial':
        # y = m[0] + m[1]*x + m[2]*x**2 +...
        y = model[0]*np.ones([npts])
        for iord in range(1,nModelParameters):
            y += model[iord]*xpts**iord
    elif basis == 'fourier':
        if nModelParameters%2==0: 
            raise ValueError("Fourier basis requires odd number of model parameters")
        if not np.all(0<= xpts) and np.all(xpts<= domainLength): 
            raise ValueError("For Fourier basis, all sample points must be in interval (0,domainLength)")
        # y = m[0]/2 + m[1] sin (pi x/L) + m[2] cos (pi x/L) + m[3] sin(pi x/L) + ...
        y = np.ones([npts])*0.5*model[0]
        n = 1
        for i in range(1,nModelParameters,2):
            y += model[i]*np.sin(2*n*np.pi*xpts/domainLength) + model[i+1]*np.cos(2*n*np.pi*xpts/domainLength)
            n+=1
    elif basis == 'discrete':
        if not np.all(0<= xpts) and np.all(xpts<= domainLength): 
            raise ValueError("For discrete basis, all sample points must be in interval (0,domainLength)")
        bounds = np.linspace(0,domainLength,nModelParameters+1)
        y = np.zeros([npts])
        for ipt in range(0,npts):
            y[ipt] = model[max(0,np.searchsorted(bounds,xpts[ipt])-1)]
    else:
        raise ValueError("Unrecognised  value for 'basis'; please specify one of: 'polynomial', 'fourier' or 'discrete'")
    if singlePoint:
        return y[0]
    else:
        return y

def curveFittingJac(xpts,nModelParameters, basis='polynomial',domainLength=1.):
    singlePoint=False
    if domainLength<0:raise ValueError("Argument 'domainLength' must be positive")
    if type(xpts) is type([]):xpts=np.array(xpts)
    try:
        npts = xpts.shape[0]
        if len(xpts.shape)!=1: raise ValueError("Argument 'xpts' should be a 1-D array")
    except AttributeError:
        singlePoint = True
        npts = 1

    if basis == 'polynomial':
        G = np.zeros([npts,nModelParameters])
        for iord in range(0,nModelParameters):
            G[:,iord] = xpts**iord
    elif basis == 'fourier':
        if nModelParameters%2==0: raise ValueError("Fourier basis requires an odd number of model parameters")
        G = np.zeros([npts,nModelParameters])
        G[:,0] = 0.5
        n = 1
        for i in range(1,nModelParameters,2):
            G[:,i] = np.sin(2*n*np.pi*xpts/domainLength)
            G[:,i+1] = np.cos(2*n*np.pi*xpts/domainLength)
            n+=1
    elif basis == 'discrete':
        G = np.zeros([npts, nModelParameters])
        bounds = np.linspace(0,domainLength,nModelParameters+1)
        y = np.zeros([npts])
        for ipt in range(0,npts):
            G[ipt,max(0,np.searchsorted(bounds,xpts[ipt])-1)] = 1.
    else:
        raise ValueError("Unsupported basis")
    return G



import numpy as np
from espresso import EspressoProblem
from espresso.exceptions import InvalidExampleError
from espresso.utils import loadtxt, absolute_path
from PIL import Image
import matplotlib.pyplot as plt
import tqdm



class XrayTracer(EspressoProblem):
    """Forward simulation class
    """

    metadata = {
        "problem_title": "X-ray Tracer",
        "problem_short_description": "",

        "author_names": ["Andrew Valentine"],

        "contact_name": "Andrew Valentine",
        "contact_email": "andrew.valentine@durham.ac.uk",

        "citations": [("Tarantola, A., 2005. *Inverse problem theory and methods for parameter estimation*, Sec.5.6.","")],

        "linked_sites": [],
    }

    def __init__(self, example_number=1):
        super().__init__(example_number)
        if example_number == 1:
            self._paths, self._attns = load_data('data/example1.dat')
            self._desc = "A straightforward X-ray tracer setup with good data coverage (InLab logo)"
            self._ngrid = 50 
            self._start = np.ones((self._ngrid,self._ngrid))
            self._true = pngToModel('data/inlab_logo.png',self._ngrid,self._ngrid,2,0.5)
        elif example_number == 2:
            self._paths, self._attns = load_data('data/example2.dat')
            self._desc = "A straightforward X-ray tracer setup with good data coverage (CSIRO logo)"
            self._ngrid = 50 
            self._start = np.ones((self._ngrid,self._ngrid))
            self._true = pngToModel('data/csiro_logo.png',self._ngrid,self._ngrid)
        elif example_number == 3:
            self._paths, self._attns = load_data('data/example3.dat')
            self._desc = "X-ray tracer with large gaps (CSIRO logo)"
            self._ngrid = 50
            self._start = np.ones((self._ngrid,self._ngrid))
            self._true = pngToModel('data/csiro_logo.png',self._ngrid,self._ngrid)
        else:
            raise InvalidExampleError

    @property
    def description(self):
        return self._desc

    @property
    def model_size(self):
        return self._ngrid**2

    @property
    def data_size(self):
        return self._attns.size

    @property
    def good_model(self):
        return self._true.flatten()

    @property
    def starting_model(self):
        return self._start.flatten()
    
    @property
    def data(self):
        return self._attns.copy()

    @property
    def covariance_matrix(self):                # optional
        raise NotImplementedError

    @property
    def inverse_covariance_matrix(self):
        raise NotImplementedError               # optional
        
    def forward(self, model, with_jacobian=False):
        n = model.size
        ngrid = int(n**0.5)
        attns,A = tracer(model.reshape((ngrid,ngrid)),self._paths)
        if with_jacobian:
            return attns, A
        else:
            return attns
    
    def jacobian(self, model):
        n = model.size
        ngrid = int(n**0.5)
        attns,A = tracer(model.reshape((ngrid,ngrid)),self._paths)
        return A

    def plot_model(self, model, paths=False, **kwargs):
        m = model.reshape((self._ngrid,self._ngrid))
        fig = plt.figure()
        ax = fig.subplots(1,1)
        im = ax.imshow(m.T,cmap=plt.cm.Blues,extent=(0,1,0,1),origin='lower',**kwargs)
        # ax.set_xticks([])
        # ax.set_yticks([])
        plt.colorbar(im,ax=ax,label='Density')
        if paths:
            for p in self._paths:
                ax.plot([p[0],p[2]],[p[1],p[3]],'y',linewidth=0.05)
        return fig
    
    def plot_data(self, data, data2=None):
        raise NotImplementedError               # optional

    def misfit(self, data, data2):              # optional
        raise NotImplementedError

    def log_likelihood(self,data,data2):        # optional
        raise NotImplementedError
    
    def log_prior(self, model):                 # optional
        raise NotImplementedError



def tracer(model,paths,extent=(0,1,0,1)):
    try:
        nx,ny = model.shape
    except:
        raise ValueError("Argument 'model' must be a 2-D numpy array")
    try:
        xmin,xmax,ymin,ymax = extent
    except:
        raise ValueError("Argument 'extent' must be a tuple,list or 1-D array with 4 elements (xmin,xmax,ymin,ymax)")
    if type(paths) == type([]):
        paths = np.array(paths)
    try:
        npaths,ncomp = paths.shape
    except:
        raise ValueError("Argument 'paths' must be a list or 2-D array")
    if ncomp!=4: raise ValueError("Each path must be described by four elements (xstart,ystart,xend,yend)")
    if any(paths[:,0]<xmin) or any(paths[:,0]>xmax) or any(paths[:,1]<ymin) or any(paths[:,1]>ymax) \
                        or any(paths[:,2]<xmin) or any(paths[:,2]>xmax) or any(paths[:,3]<ymin) or any(paths[:,3]>ymax):
        raise ValueError("All sources and receivers must be within or on boundary of model region")

    xGridBounds = np.linspace(xmin,xmax,nx+1)
    yGridBounds = np.linspace(ymin,ymax,ny+1)
    A = np.zeros([npaths,nx*ny])
    attns = np.zeros([npaths])
    t = tqdm.tqdm(desc="Evaluating paths",total=npaths)
    for ip,p in enumerate(paths):
        xs,ys,xr,yr = p
        pathLength = np.sqrt((xr-xs)**2 + (yr-ys)**2)
        # Compute lambda for intersection with each grid-line
        lamX = np.array([]) if xr==xs else np.array([(d-xs)/(xr-xs) for d in xGridBounds])
        lamY = np.array([]) if yr==ys else np.array([(d-ys)/(yr-ys) for d in yGridBounds])
        # Discard any intersections that would occur outside the region
        lamX = np.extract(np.logical_and(lamX>=xmin,lamX<=xmax),lamX)
        nlamX = len(lamX)
        lamY = np.extract(np.logical_and(lamY>=ymin,lamY<=ymax),lamY)
        lam = np.concatenate( (lamX,lamY) )
        lamSort = np.argsort(lam,kind='mergesort')
        dx = 1 if xr>xs else -1
        dy = 1 if yr>ys else -1
        try:
            if lam[lamSort[0]]!=0:
                lam = np.concatenate((np.array([0]),lam))
                lamSort = np.concatenate((np.array([0]),lamSort+1))
                nlamX+=1
        except IndexError:
            lam = np.array([0])
            lamSort = np.array([0])
        if lam[lamSort[-1]]!=1:
            lam = np.concatenate((lam,np.array([1])))
            lamSort = np.concatenate((lamSort,np.array([lam.shape[0]-1])))
        if xs==xmin:
            ix = 0
        elif xs==xmax:
            ix = nx-1
        else:
            ix = np.searchsorted(xGridBounds,xs,side='right' if xr>xs else 'left')-1
        if ys==ymin:
            iy = 0
        elif ys==ymax:
            iy = ny-1
        else:
            iy = np.searchsorted(yGridBounds,ys,side='right' if yr>ys else 'left')-1
        pathSensitivity = np.zeros_like(model)
        ilam0=2 if lam[lamSort[1]]==0 else 1
        for ilam in range(ilam0,len(lam)):
            dl = (lam[lamSort[ilam]] - lam[lamSort[ilam-1]])*pathLength
            pathSensitivity[ix,iy] = dl
            attns[ip]+=dl*model[ix,iy]
            if lamSort[ilam]>=nlamX:
                iy+=dy
            else:
                ix+=dx
            if lam[lamSort[ilam]]==1.:break

        A[ip,:] = pathSensitivity.flatten()
        t.update(1)
    t.close()

    return attns,A

def load_data(filename):
    data = loadtxt(filename,skiprows=1)
    ndata = data.shape[0]
    paths = np.zeros((ndata,4))
    attns = np.zeros(ndata)
    paths[:,0:2] = data[:,0:2]
    paths[:,2:4] = data[:,3:5]
    attns = -np.log(data[:,5]/data[:,2])
    return paths, attns

def generateExampleDataset(img_filename, out_filename):
    noiseLevels=None #[0.005,0.01,0.015,0.02,0.025]
    noiseLevels=[0.0001]
    # m = pngToModel(img_filename,1024,1024,1,1)
    m = pngToModel(img_filename,1024,1024,2,0.5)
    # srcs = np.array([[0,0],[0,0.2],[0.,0.4],[0,0.5],[0,0.6],[0.,0.65],[0.,0.7]]+[[0,x] for x in np.linspace(0.71,1.0,30)]+[[x,0] for x in np.linspace(0.3,0.6,30)])
    # recs = generateSurfacePoints(40,surface=[False,True,False,True])
    # recs = generateSurfacePoints(50,surface=[False,True,False,False])
    # srcs = generateSurfacePoints(50,surface=[False,False,True,False])
    recs = generateSurfacePoints(30,surface=[True,True,True,True])
    srcs = generateSurfacePoints(20,surface=[True,True,True,True])
    paths = buildPaths(srcs,recs)
    Isrc = np.random.uniform(10,1,size=paths.shape[0])
    attns,A = tracer(m,paths)
    Irec = Isrc*np.exp(-attns)
    if noiseLevels is not None:
        noise = np.zeros([paths.shape[0]])
        for i in range(0,paths.shape[0]):
            noise[i] = np.random.choice(noiseLevels)
            Irec[i]+=np.random.normal(0,noise[i])
            if Irec[i]<=0: Irec[i] = 1.e-3

    fp = open(out_filename,'w')
    fp.write("# Src-x Src-y Src-Int Rec-x Rec-y Rec-Int")
    if noiseLevels is None:
        fp.write("\n")
    else:
        fp.write(" Rec-sig\n")
    for i in range(0,paths.shape[0]):
        fp.write("%2.4f %2.4f %2.4f %2.4f %2.4f %2.4f"%(paths[i,0],paths[i,1],Isrc[i],paths[i,2],paths[i,3],Irec[i]))
        if noiseLevels is None:
            fp.write("\n")
        else:
            fp.write(" %2.4f\n"%noise[i])
    fp.close()

def buildPaths(srcs,recs):
    if type(srcs) is type([]): srcs = np.array(srcs)
    try:
        nsrcs,nc = srcs.shape
    except:
        raise ValueError("Argument 'srcs' must be a 2-D nummpy array")
    if nc!=2: raise ValueError("Argument 'srcs' should have shape (N x 2)")
    if type(recs) is type([]): recs = np.array(recs)
    try:
        nrecs,nc = recs.shape
    except:
        raise ValueError("Argument 'recs' must be a 2-D nummpy array")
    if nc!=2: raise ValueError("Argument 'recs' should have shape (N x 2)")
    npaths = nsrcs*nrecs
    paths = np.zeros([npaths,4])
    ip=0
    for isrc in range(nsrcs):
        for irec in range(nrecs):
            paths[ip,0:2] = srcs[isrc,:]
            paths[ip,2:4] = recs[irec,:]
            ip+=1
    return paths

def generateSurfacePoints(nPerSide,extent=(0,1,0,1),surface=[True,True,True,True],addCorners=True):
    out = []
    if surface[0]:
        out+=[[extent[0],x] for x in np.linspace(extent[2],extent[3],nPerSide+2)[1:nPerSide+1]]
    if surface[1]:
        out+=[[extent[1],x] for x in np.linspace(extent[2],extent[3],nPerSide+2)[1:nPerSide+1]]
    if surface[2]:
        out+=[[x,extent[2]] for x in np.linspace(extent[0],extent[1],nPerSide+2)[1:nPerSide+1]]
    if surface[3]:
        out+=[[x,extent[3]] for x in np.linspace(extent[0],extent[1],nPerSide+2)[1:nPerSide+1]]
    if addCorners:
        if surface[0] or surface[2]:
            out+=[[extent[0],extent[2]]]
        if surface[0] or surface[3]:
            out+=[[extent[0],extent[3]]]
        if surface[1] or surface[2]:
            out+=[[extent[1],extent[2]]]
        if surface[1] or surface[3]:
            out+=[[extent[1],extent[3]]]
    return np.array(out)

def pngToModel(pngfile,nx,ny,bg=1.,sc=1.):
    png = Image.open(absolute_path(pngfile))
    png.load()

    try:
        model = sc*(bg+np.asarray(png.convert('L').resize((nx,ny)).transpose(Image.Transpose.ROTATE_270))/255.)
    except:
        model = sc*(bg+np.asarray(png.convert('L').resize((nx,ny)).transpose(Image.ROTATE_270))/255.)        
    return model


import os, errno


def silent_remove(filename):
    r"""Remove a file if exists or do nothing otherwise

    Parameters
    ----------
    filename : Path | str
        The absoluate path of the file you want to remove
    """
    try:
        os.remove(filename)
    except OSError as e:
        if e.errno != errno.ENOENT: # errno.ENOENT = no such file or directory
            raise # re-raise exception if a different error occurred



"""Module initialisation file

Code inside this file will be executed when this submodule is imported.

For contributors: feel free to add your initialisation code here if needed.
"""

from ._slug_test import *

__all__ = [ "SlugTest" ]



from espresso import EspressoProblem
from espresso.exceptions import InvalidExampleError


class ExampleName(EspressoProblem):
    """Forward simulation class"""

    # TODO fill in the following metadata.
    metadata = {
        "problem_title": "",  # To be used in docs
        "problem_short_description": "",  # 1-3 sentences
        # List of names e.g. author_names = ["Sally Smith", "Mark Brown"]
        "author_names": [
            "",
            "",
        ],
        # Contact for contributor/maintainer of espresso example
        "contact_name": "",
        "contact_email": "",
        # Reference to publication(s) that describe this example. In most
        # cases there will only be a single entry in this list.
        # List of (citation, doi) pairs e.g.
        # citations = [("Newton, I (1687). Philosophiae naturalis principia mathematica.", "")]
        # If there are no citations, use empty list `[]`
        "citations": [("", "")],
        # List of (title, address) pairs for any websites that
        # should be linked in the documentation, e.g.
        # linked_sites = [("Parent project on Github","https://github.com/user/repo"),
        #                 ("Data source"),"https://www.data.com") ]
        # If there are no links, use empty list `[]`
        "linked_sites": [("", "")],
    }

    def __init__(self, example_number=1):
        super().__init__(example_number)

        """you might want to set some useful example-specific parameters here
        """
        # if example_number == 1:
        #     self.some_attribute = some_value_0
        #     self.another_attribute = another_value_0
        # elif example_number == 2:
        #     self.some_attribute = some_value_1
        #     self.another_attribute = another_value_1
        # else:
        #     raise InvalidExampleError

    @property
    def description(self):
        raise NotImplementedError  # optional

    @property
    def model_size(self):
        raise NotImplementedError  # TODO implement me

    @property
    def data_size(self):
        raise NotImplementedError  # TODO implement me

    @property
    def good_model(self):
        raise NotImplementedError  # TODO implement me

    @property
    def starting_model(self):
        raise NotImplementedError  # TODO implement me

    @property
    def data(self):
        raise NotImplementedError  # TODO implement me

    @property
    def covariance_matrix(self):  # optional
        raise NotImplementedError

    @property
    def inverse_covariance_matrix(self):
        raise NotImplementedError  # optional

    def forward(self, model, with_jacobian=False):
        if with_jacobian:
            raise NotImplementedError  # optional
        else:
            raise NotImplementedError  # TODO implement me

    def jacobian(self, model):
        raise NotImplementedError  # optional

    def plot_model(self, model):
        raise NotImplementedError  # optional

    def plot_data(self, data, data2=None):
        raise NotImplementedError  # optional

    def misfit(self, data, data2):
        raise NotImplementedError  # optional

    def log_likelihood(self, data, data2):
        raise NotImplementedError  # optional

    def log_prior(self, model):
        raise NotImplementedError  # optional



from pathlib import Path
import inspect
import numpy as np


def _absolute_path(relative_path):
    # get caller file path
    caller_frame = inspect.stack()[2]
    caller_filename_full = caller_frame.filename
    # combine path
    file_path = (Path(caller_filename_full).parent / relative_path).resolve()
    return file_path

def absolute_path(relative_path):
    r"""Returns the absolute path of a data file

    Please always use this function or :func:`loadtxt` when trying to load data from 
    a relative path into an Espresso problem class.

    Parameters
    ----------
    relative_path : str
        The relative path of the file you want the absoluate path for

    Examples
    --------

    >>> from espresso.utils import absolute_path
    >>> from PIL import Image
    >>> png = Image.open(absolute_path("data/csiro_logo.png"))

    """
    return _absolute_path(relative_path)

def loadtxt(relative_path, *args, **kwargs):
    r"""Wrapper of :func:`numpy.loadtxt` given a relative path

    Please always use this function or :func:`absolute_path` when trying to load data 
    from a relative path into an Espresso problem class.

    Parameters
    ----------
    relative_path : str
        The relative path of the file you want to load
    *args : any
        Passed directly into :func:`numpy.loadtxt`
    **kwargs : any
        Passed directly into :func:`numpy.loadtxt`

    Examples
    --------

    >>> from espresso.utils import loadtxt
    >>> data = loadtxt("data/example1.dat")
    >>> data
    array([[0.    , 0.0323, 5.9987, 1.    , 0.0323, 2.057 ],
        [0.    , 0.0323, 6.3879, 1.    , 0.0645, 2.1543],
        [0.    , 0.0323, 1.8923, 1.    , 0.0968, 0.6279],
        ...,
        [1.    , 0.    , 4.1087, 0.    , 1.    , 0.5583],
        [1.    , 0.    , 3.9118, 1.    , 0.    , 3.9118],
        [1.    , 0.    , 9.2198, 1.    , 1.    , 3.2873]])

    """
    abs_path = _absolute_path(relative_path)
    return np.loadtxt(abs_path, *args, **kwargs)



# Command to create a new example folder:
# In ROOT, execute:
# python espresso_machine/new_contribution/create_new_contrib.py <example-name>
# Replacing <example_name> with the new example name.

from glob import glob
import sys
import os
from shutil import copyfile
from pathlib import Path


def getListOfFiles(dirName):
    # create a list of file and sub directories
    # names in the given directory
    listOfFile = os.listdir(dirName)
    allFiles = list()
    # Iterate over all the entries
    for entry in listOfFile:
        # Create full path
        fullPath = os.path.join(dirName, entry)
        # If entry is a directory then get the list of files in this directory
        if os.path.isdir(fullPath):
            allFiles = allFiles + getListOfFiles(fullPath)
        else:
            allFiles.append(fullPath)

    return allFiles


current_dir = Path(__file__).resolve().parent
root_dir = current_dir.parent.parent
CONTRIB_FOLDER = str(root_dir / "contrib")
TEMPLATE_FOLDER = str(root_dir / "espresso_machine/new_contribution/_template")


def main():
    print(
        "🥰 Thanks for contributing! \nWe are generating new contribution component from"
        " template...\n"
    )

    # validate example name
    if len(sys.argv) != 2:
        raise RuntimeError(
            "No example name detected.\n\nUsage: python create_new_contrib.py"
            " EXAMPLE_NAME\n\n"
        )
    example_name = sys.argv[-1]
    existing_examples = glob(CONTRIB_FOLDER + "/*/")
    existing_examples = [e.split("/")[-2] for e in existing_examples]
    if example_name in existing_examples:
        raise ValueError(
            "The example name provided already exists, please choose another name"
        )
    elif example_name in ["utils", "_machine"]:
        raise ValueError(
            "This sub-folder name is occupied in Espresso core library, "
            "please choose another name"
        )

    # convert example name to other formats
    example_name_capitalised = example_name.title().replace("_", "").replace("-", "")

    # make new folders and subfolders
    new_subfolder = CONTRIB_FOLDER + "/" + example_name
    os.makedirs(new_subfolder)
    os.makedirs(new_subfolder + "/data")

    # generate file names
    template_files = getListOfFiles(TEMPLATE_FOLDER)
    new_files = [f.replace(TEMPLATE_FOLDER, new_subfolder) for f in template_files]
    new_files = [f.replace("example_name", example_name) for f in new_files]

    # generate example subfolder from template
    print("🗂  Copying files...")
    for template_file, new_file in zip(template_files, new_files):
        print("file: " + template_file + " -> " + new_file)
        files_to_adapt = ["README", "__init__.py", "example_name.py"]
        if any([fname in template_file for fname in files_to_adapt]):
            with open(template_file, "r") as template_f:
                content = template_f.read()
            content = content.replace("example_name", example_name)
            content = content.replace("Example Name Title", example_name_capitalised)
            content = content.replace("ExampleName", example_name_no_space)
            with open(new_file, "w") as new_f:
                new_f.write(content)
        else:
            copyfile(template_file, new_file)
    print(
        "\n🎉 OK. Please navigate to " + new_subfolder + " to write your own example. "
    )


if __name__ == "__main__":
    main()



"""Build the Python package "espresso"

1. clean "_esp_build/"
2. "/<meta-data-files>" => "_esp_build/"
3. generate "_version.py"
4. "src/" => "_esp_build/src/"
5. "contrib/" => "_esp_build/src/espresso/" + "__init__.py" + "list_problems.py"
6. remove `.core` from versioningit_config
7. "espresso_machine/" => _esp_build/src/_machine"
8. build capability_matrix
9. `pip install .`      (can be disabled by `--no-install`)

Usage: python build.py [--pre] [--post] [--no-install] [-c <example_name>] [--file <file_name>]
"""

import subprocess
import sys
import os
from shutil import copytree, copy, rmtree, ignore_patterns
from pathlib import Path
import versioningit
import json

import _utils
import report
import validate


# ------------------------ constants ------------------------
MODULE_NAME = "espresso"
PKG_NAME = "geo-espresso"
current_directory = Path(__file__).resolve().parent
root = current_directory.parent.parent
ROOT_DIR = str(root)
BUILD_DIR = str(root / "_esp_build")
PKG_SRC = str(root / "src")
CONTRIB_SRC = str(root / "contrib")
VCS_GIT = str(root / ".git")
DOCS_SRC = str(root / "docs")
MACHINE_SRC = str(root / "espresso_machine")
META_FILES = [
    "README.md",
    "pyproject.toml",
    "setup.py",
    "LICENCE",
    "CMakeLists.txt",
    ".readthedocs.yml",
    ".gitignore",
    "CHANGELOG.md",
]
PROBLEMS_TO_COMPILE_FILE = "problems_to_compile.txt"
validate_script = str(Path(__file__).resolve().parent / "validate.py")


# ------------------------ helpers ------------------------
def is_cache(file_name):
    return (
        file_name.endswith(".pyc")
        or file_name == "__pycache__"
        or file_name == "cmake_install.cmake"
        or file_name.endswith(".mod")
        or file_name.endswith(".out")
        or file_name == "CMakeFiles"
        or file_name == "Makefile"
        or file_name == PROBLEMS_TO_COMPILE_FILE
    )


def move_folder_content(folder_path, dest_path, prefix=None, only_include=None):
    if prefix is None:
        copytree(
            folder_path,
            dest_path,
            # dirs_exist_ok=True,
            ignore=ignore_patterns("*.pyc", "tmp*", "__pycache__"),
        )
    else:           # moving contributions source
        for f in os.listdir(folder_path):
            src = f"{folder_path}/{f}"
            dst = f"{dest_path}/{prefix}{f}"
            if is_cache(f) or not os.path.isdir(src) or (only_include is not None and f not in only_include):
                continue
            copytree(
                src,
                dst,
                ignore=ignore_patterns(
                    "*.pyc",
                    "__pycache__",
                    "tmp*",
                    "CMakeFiles",
                    "Makefile",
                    "*.mod",
                    "*.out",
                ),
            )
            # add underscore prefix to file name
            for ff in os.listdir(dst):
                if ff == f"{f}.py":
                    ff_origin = f"{dst}/{ff}"
                    ff_rename = f"{dst}/{prefix}{ff}"
                    os.rename(ff_origin, ff_rename)
                if ff == "__init__.py" or ff == "CMakeLists.txt":
                    with open(f"{dst}/{ff}", "r") as fff:
                        lines = fff.readlines()
                    with open(f"{dst}/{ff}", "w") as fff:
                        for line in lines:
                            fff.write(line.replace(f, f"_{f}"))


# ------------------------ tasks ------------------------
# 1
def clean_build_folder():
    dirpath = Path(BUILD_DIR)
    if dirpath.exists() and dirpath.is_dir():
        rmtree(dirpath)


# 2
def move_pkg_metadata():
    move_folder_content(DOCS_SRC, f"{BUILD_DIR}/docs")
    for f in META_FILES:
        copy(f"{ROOT_DIR}/{f}", f"{BUILD_DIR}/{f}")


# 3
def write_version():
    versioningit_config = {
        "format": {
            "distance": "{base_version}+{distance}.{vcs}{rev}",
            "dirty": "{base_version}+{distance}.{vcs}{rev}.dirty",
            "distance-dirty": "{base_version}+{distance}.{vcs}{rev}.dirty",
        },
        "write": {"file": "src/espresso/_version.py"},
    }
    versioningit.get_version(root, versioningit_config, True)


# 4
def move_pkg_source():
    move_folder_content(PKG_SRC, f"{BUILD_DIR}/src")


# 5
def change_versioningit_config():
    with open(f"{BUILD_DIR}/setup.py", "r") as f:
        setup_content = f.read()
    setup_content = setup_content.replace(".core", "")
    setup_content = setup_content.replace('"rmsuffix": ""', '"rmsuffix": "-build"')
    with open(f"{BUILD_DIR}/setup.py", "w") as f:
        f.write(setup_content)


# 6
def move_contrib_source():
    # see if any contribution is specified through command line args
    specified_problems = _utils.problems_to_run_names_only()
    # move all contribution subfolders with prefix "_"
    move_folder_content(
        CONTRIB_SRC,
        f"{BUILD_DIR}/src/{MODULE_NAME}",
        prefix="_",
        only_include=specified_problems,
    )
    # collect a list of contributions + related strings to write later
    contribs = []
    init_file_imports = "\n"
    init_file_all_cls = "\n_all_problems = [\n"
    for path in Path(CONTRIB_SRC).iterdir():
        contrib = os.path.basename(path)  # name
        if path.is_dir() and (
            specified_problems is None or contrib in specified_problems
        ):
            contrib_class = _utils.problem_name_to_class(contrib)  # class
            contribs.append(contrib)
            init_file_imports += f"from ._{contrib} import {contrib_class}\n"
            init_file_all_cls += f"    {contrib_class},\n"
    init_file_all_cls += "]"
    # some constant strings to append to init file later
    init_file_imp_funcs = (
        "\nfrom .list_problems import list_problem_names, list_problems"
    )
    init_file_add_all_nms = "\n__all__ += list_problem_names()"
    init_file_add_funcs = "\n__all__ += ['list_problem_names', 'list_problems']\n"
    # write all above to files
    # compiled_code_list = set()
    with open(f"{BUILD_DIR}/src/{MODULE_NAME}/CMakeLists.txt", "a") as f:
        for contrib in contribs:
            f.write(f"install(DIRECTORY _{contrib} DESTINATION .)\n")
            if Path(f"{CONTRIB_SRC}/{contrib}/CMakeLists.txt").exists():
                f.write(f"add_subdirectory(_{contrib})\n")
                # compiled_code_list.add(contrib)
    with open(f"{BUILD_DIR}/src/{MODULE_NAME}/__init__.py", "a") as f:
        f.write(init_file_imports)
        f.write(init_file_imp_funcs)
        f.write(init_file_add_all_nms)
        f.write(init_file_add_funcs)
    with open(f"{BUILD_DIR}/src/{MODULE_NAME}/list_problems.py", "a") as f:
        f.write(init_file_imports)
        f.write(init_file_all_cls)
    # with open(f"{ROOT_DIR}/contrib/{PROBLEMS_TO_COMPILE_FILE}", "w") as f:
    #     f.writelines(compiled_code_list)


# 7 move espresso_machine into espresso/_machine
def move_espresso_machine():
    move_folder_content(MACHINE_SRC, f"{BUILD_DIR}/src/espresso/_machine")


# 8 build capability matrix
def build_problem_capability():
    # see if any contribution is specified through command line args
    specified_problems = _utils.problems_to_run_names_only()
    with _utils.suppress_stdout():
        capability_report = report.capability_report(
            problems_to_check=specified_problems
        )
    report_to_write = json.dumps(capability_report, indent=4)
    with open(f"{BUILD_DIR}/src/{MODULE_NAME}/list_problems.py", "a") as f:
        f.write("\n\n_capability_matrix = ")
        f.write(report_to_write)


# printing helper
def println_with_emoji(content, emoji):
    try:
        print(f"\n{emoji} {content}")
    except:
        print(f"\n{content}")


# ------------------------ main functions ------------------------
build_pipeline = [
    (clean_build_folder, "Cleaning build folder..."),
    (move_pkg_metadata, "Moving package metadata..."),
    (write_version, "Generating version file..."),
    (move_pkg_source, "Moving Espresso core packaging files..."),
    (change_versioningit_config, "Removing `.core` from versioningit config..."),
    (move_contrib_source, "Moving all contributions..."),
    (move_espresso_machine, "Moving infrastructure code..."),
    (
        build_problem_capability,
        "Building capability matrix... (this will take some time)",
    ),
    # ( install_pkg, "Building Python package: geo-espresso..." ),
]


def build():
    println_with_emoji("Package building...", "🛠")
    for (step, desc) in build_pipeline:
        println_with_emoji(desc, "🗂")
        try:
            step()
        except Exception as e:
            println_with_emoji(f"Problem occurred in this step: {desc}\n", "❗️")
            raise e
        print("OK.")
    # println_with_emoji("Espresso installed!", "🍰")


def install_pkg():
    desc = "Building Python package: geo-espresso..."
    println_with_emoji(desc, "🗂")
    subprocess.call([sys.executable, "-m", "pip", "uninstall", "-y", PKG_NAME])
    res = subprocess.call([sys.executable, "-m", "pip", "install", "."], cwd=BUILD_DIR)
    if res != 0:
        sys.exit(res)


def pre_validate():
    validate.main(pre_build=True)


def post_validate():
    validate.main(pre_build=False)


def main():
    _args = _utils.args()
    if _args.pre:
        pre_validate()
    build()
    if not _args.dont_install:
        install_pkg()
    if _args.post:
        post_validate()
    println_with_emoji("All done", "🍰")


if __name__ == "__main__":
    main()



"""Module initialisation file

It's NOT recommended to put extra code inside this file. Code inside this file will be
executed when this submodule is imported, so adding things in this file can slow down
the importing process of `espresso`.

For contributors, add any intialisation code for your problem into _magnetotelluric_1D.py, 
under the method `__init__()` of the class `Magnetotelluric1D`.

Don't touch this file unless you know what you are doing :)
"""

from ._magnetotelluric_1D import *

__all__ = [ "Magnetotelluric1D" ]



"""Helper functions to analyse the base class `EspressoProblem`

This script assumes you have geo-espresso installed via:
    $ python espresso_machine/build_package/build.py
Or the core package via:
    $ pip install .

The following two functions are exposed to other scripts by this file:
- criteria_for_problems
- criteria_for_example

Check comments for these functions for what criteria are implemented.
"""

import os
import numpy as np
from matplotlib.figure import Figure

import run_examples
import _utils

try:
    import espresso
except ModuleNotFoundError as e:
    e.msg += (
        "\n\nNote: To run pre-build validation, please firstly install "
        "`espresso` core module by running the following from the root"
        "level of the project\n  $ pip install ."
    )
    raise e


# --------------> validate a contributed problem

# 1. Contribution folder name matches the main Python file name
def _check_folder_file_names(prob_name, prob_path, names_in_folder):
    assert f"{prob_name}.py" in names_in_folder, (
        "the contribution folder name should match the main Python file name, "
        f"so you should have this file: contrib/{prob_name}/{prob_name}.py"
    )


# 2. These file exist: README.md, LICENCE, __init__.py
def _check_required_files(prob_path, names_in_folder):
    required_files = ["README.md", "LICENCE", "__init__.py"]
    for file in required_files:
        assert (
            file in names_in_folder
        ), f"{file} is required but you don't have it in {prob_path}"


# 3. LICENCE file is not empty
def _check_licence_nonempty(prob_path):
    assert (
        os.stat(f"{prob_path}/LICENCE").st_size != 0
    ), "ensure the LICENCE file is not empty"


# 4. The __init__.py contains `__all__` variable and has the class name in it
def _check_init_all(parent_mod, prob_name, prob_class_name):
    _class_not_in_init_all_msg = (
        "make sure you include "
        f"`__all__ = ['{prob_class_name}']` in the file "
        f"contrib/{prob_name}/__init__.py"
    )
    try:
        _init_all = parent_mod.__all__
    except:
        raise AssertionError(_class_not_in_init_all_msg)
    else:
        assert prob_class_name in _init_all, (
            _class_not_in_init_all_msg
            + f". We found {_init_all} but {prob_class_name} is not in there"
        )


# 5. The class is a subclass of EspressoProblem
def _check_subclass(prob_class, prob_class_name):
    assert issubclass(prob_class, espresso.EspressoProblem), (
        f"make sure your problem class `{prob_class_name}` is a subclass of "
        "`espresso.EspressoProblem`, by defining it with:\n\n-------\n"
        "from espresso import EspressoProblem\n\n"
        f"class {prob_class_name}(EspressoProblem):\n"
        "\tdef __init__(self, example_number=1):\n"
        "\t\tsuper().__init__(example_number)\n\t\t...\n\n-------"
    )


# 6. The following metadata are defined properly:
#     - problem_title
#     - problem_short_description
#     - author_names
#     - contact_name
#     - contact_email
#     - [optional] citations -> []
#     - [optional] linked_sites -> [(name, link)]
def _check_metadata(prob_class, prob_class_name):
    class_metadata = prob_class.metadata
    # problem_title
    assert (
        type(class_metadata["problem_title"]) is str
        and len(class_metadata["problem_title"]) > 0
    ), (
        f"check class attribute `{prob_class_name}.metadata['problem_title']`"
        " is present and is a non-empty string"
    )
    # problem_short_description
    assert (
        type(class_metadata["problem_short_description"]) is str
    ), (  # Allow empty field
        "check class attribute "
        f"`{prob_class_name}.metadata['problem_short_description']`"
        " is present and is a string"
    )
    # author_names
    assert type(class_metadata["author_names"]) is list, (
        "check class attribute "
        f"`{prob_class_name}.metadata['author_names']`"
        " is present and is a list"
    )
    assert len(class_metadata["author_names"]) > 0, (
        "check class attribute "
        f"`{prob_class_name}.metadata['author_names']` is not empty"
    )
    for author in class_metadata["author_names"]:
        assert type(author) is str and len(author) > 0, (
            "check elements of class attribute "
            f"`{prob_class_name}.metadata['author_names`] are non-empty strings"
        )
    # contact_name
    assert (
        type(class_metadata["contact_name"]) is str
        and len(class_metadata["contact_name"]) > 0
    ), (
        "check class attribute "
        f"`{prob_class_name}.metadata['contact_name']`"
        " is present and is a non-empty string"
    )
    # contact_email
    assert (
        type(class_metadata["contact_email"]) is str
        and "@" in class_metadata["contact_email"]
    ), (
        "check class attribute "
        f"`{prob_class_name}.metadata['contact_email']`"
        " is present and is a valid email address string"
    )
    # citations
    assert type(class_metadata["citations"]) is list, (
        "check class attribute "
        f"`{prob_class_name}.metadata['citations']`"
        " is present and is a list"
    )
    for citation in class_metadata["citations"]:
        assert type(citation) is tuple and len(citation) == 2, (
            "check elements of class attribute "
            f"`{prob_class_name}.metadata['citations`] are 2-element tuples"
        )
        for field in citation:
            assert type(field) is str, (
                "check elements of class attribute "
                f"`{prob_class_name}.metadata['citations`] have tuples of "
                "strings"
            )
    # linked_sites
    assert type(class_metadata["linked_sites"]) is list, (
        "check class attribute "
        f"`{prob_class_name}.metadata['linked_sites']`"
        " is present and is a list"
    )
    for site in class_metadata["linked_sites"]:
        assert type(site) is tuple and len(site) == 2, (
            "check elements of class attribute "
            f"`{prob_class_name}.metadata['linked_sites`] are 2-element tuples"
        )
        for field in site:
            assert type(field) is str, (
                "check elements of class attribute "
                f"`{prob_class_name}.metadata['linked_sites`] have tuples of "
                "strings"
            )


def criteria_for_problem(prob_class, prob_class_name, prob_path, parent_mod):
    """validate a contributed problem

    This function checks the folder structure and metadata of specified problem:
    1. Contribution folder name matches the main Python file name
    2. These file exist: README.md, LICENCE, __init__.py
    3. LICENCE file is not empty
    4. The __init__.py contains `__all__` variable and has the class name in it
    5. The class is a subclass of EspressoProblem
    6. The following metadata are defined properly:
        - problem_title
        - problem_short_description
        - author_names
        - contact_name
        - contact_email
        - [optional] citations -> []
        - [optional] linked_sites -> [(name, link)]

    Note: a problem will have several examples, but we will put all the criteria for
    the examples to function `criteria_for_example`.
    """
    # preparation
    names_in_folder, paths_in_folder = _utils.get_folder_content(prob_path)
    prob_name = prob_path.split("/")[-1]

    # checking
    _check_folder_file_names(prob_name, prob_path, names_in_folder)
    _check_required_files(prob_path, names_in_folder)
    _check_licence_nonempty(prob_path)
    _check_init_all(parent_mod, prob_name, prob_class_name)
    _check_subclass(prob_class, prob_class_name)
    _check_metadata(prob_class, prob_class_name)


# --------------> validate an example of problem


def _inspect_espresso_problem():
    all_props = dir(espresso.EspressoProblem)
    abs_meta = espresso.EspressoProblem.__abstract_metadata_keys__
    abs_props = espresso.EspressoProblem.__abstractmethods__
    opt_props = [p for p in all_props if p not in abs_meta and p not in abs_props]
    return {
        "required meta keys": abs_meta,
        "required attributes": abs_props,
        "optional attributes": opt_props,
    }


def _check_is_number(all_results: run_examples.ResultsFromExample, obj, obj_str):
    assert isinstance(obj, (float, int)), f"ensure {obj_str} is a number"


def _check_is_str(all_results: run_examples.ResultsFromExample, obj, obj_str):
    assert isinstance(obj, str), f"ensure {obj_str} is a string"


def _check_is_figure(all_results: run_examples.ResultsFromExample, obj, obj_str):
    assert isinstance(
        obj, Figure
    ), f"ensure {obj_str} returns an instance of matplotlib.figure.Figure"


def _check_1d_array_like(all_results: run_examples.ResultsFromExample, obj, obj_str):
    assert np.ndim(obj) == 1, f"ensure {obj_str} is a flat array"


def _check_2d_array_like(all_results: run_examples.ResultsFromExample, obj, obj_str):
    assert np.ndim(obj) == 2, f"ensure {obj_str} is a 2d array"


def _check_shape_factory(get_shape, get_shape_str):
    def _check_shape(all_results: run_examples.ResultsFromExample, obj, obj_str):
        expected_shape = get_shape(all_results)
        expected_shape_str = get_shape_str(all_results)
        assert (
            np.shape(obj) == expected_shape
        ), f"ensure {obj_str} has shape {expected_shape_str}, i.e. {expected_shape}"

    return _check_shape


def _check_shape_config_models():
    get_shape = lambda all_results: (all_results.nmodel,)
    get_shape_str = lambda all_results: f"({all_results.prob_instance_str}.model_size,)"
    return get_shape, get_shape_str


_check_model_shape = _check_shape_factory(*_check_shape_config_models())


def _check_shape_config_data():
    get_shape = lambda all_results: (all_results.ndata,)
    get_shape_str = lambda all_results: f"({all_results.prob_instance_str}.data_size,)"
    return get_shape, get_shape_str


_check_data_shape = _check_shape_factory(*_check_shape_config_data())


def _check_shape_config_cov():
    get_shape = lambda all_results: (all_results.ndata, all_results.ndata)
    get_shape_str = (
        lambda all_results: f"({all_results.prob_instance_str}.data_size, {all_results.prob_instance_str}.data_size)"
    )
    return get_shape, get_shape_str


_check_cov_shape = _check_shape_factory(*_check_shape_config_cov())


def _check_shape_config_jac():
    get_shape = lambda all_results: (all_results.ndata, all_results.nmodel)
    get_shape_str = (
        lambda all_results: f"({all_results.prob_instance_str}.data_size, {all_results.prob_instance_str}.model_size)"
    )
    return get_shape, get_shape_str


_check_jac_shape = _check_shape_factory(*_check_shape_config_jac())

attributes_to_check = [
    # ( output dict key,
    #   attribute str,
    #   whether required,
    #   how to validate given the attribute value/result and str )
    ("nmodel", "model_size", True, [_check_is_number]),
    ("ndata", "data_size", True, [_check_is_number]),
    ("model", "good_model", True, [_check_1d_array_like, _check_model_shape]),
    ("null_model", "starting_model", True, [_check_model_shape]),
    ("data", "data", True, [_check_1d_array_like, _check_data_shape]),
    ("description", "description", False, [_check_is_str]),
    ("cov", "covariance_matrix", False, [_check_2d_array_like, _check_cov_shape]),
    (
        "inv_cov",
        "inverse_covariance_matrix",
        False,
        [_check_2d_array_like, _check_cov_shape],
    ),
    (
        "synth1",
        "forward(prob.good_model)",
        True,
        [_check_1d_array_like, _check_data_shape],
    ),
    (
        "jac1",
        "jacobian(prob.good_model)",
        False,
        [_check_2d_array_like, _check_jac_shape],
    ),
    (
        "synth2",
        "forward(prob.good_model, True)[0]",
        False,
        [_check_1d_array_like, _check_data_shape],
    ),
    (
        "jac2",
        "forward(prob.good_model, True)[1]",
        False,
        [_check_2d_array_like, _check_jac_shape],
    ),
    ("fig_model", "plot_model(prob.good_model)", False, [_check_is_figure]),
    ("fig_data", "plot_data(prob.data)", False, [_check_is_figure]),
    ("misfit", "misfit(prob.data, prob.data)", False, [_check_is_number]),
    (
        "log_likelihood",
        "log_likelihood(prob.data, prob.data)",
        False,
        [_check_is_number],
    ),
    ("log_prior", "log_prior(prob.good_model)", False, [_check_is_number]),
]


def criteria_for_example(all_results):
    """validate an example under a specific problem

    This function checks that the required attributes are defined, and that if anything
    is defined then it's done properly (including the types and shapes):
    1. Required methods/properties are implemented and can run
    2. Optional methods/properties are either not implemented or have no error
    3. Any methods/properties, if implemented, are done propertly
    """
    for attr_check in attributes_to_check:
        attr_key, attr_str, required, to_check = attr_check
        obj = all_results[attr_key]
        obj_str = f"{all_results['prob_instance_str']}.{attr_str}"
        if isinstance(obj, Exception):
            raise obj
        if obj is None and required:
            raise NotImplementedError(
                f"{obj_str} is required but you haven't implemented it"
            )
        if obj is not None:
            for check_func in to_check:
                check_func(all_results, obj, obj_str)


def main():
    problems = _utils.problems_to_run(problems_specified=["testtest"])
    results = run_examples.run_problems(problems, pre_build=True, timeout=_utils.DEFAULT_TIMEOUT)
    for res in results:
        criteria_for_problem(
            res.problem_class,
            res.problem_class_str,
            res.problem_path,
            res.parent_module,
        )
        for prob_out_i in res.problem_results_generator:
            criteria_for_example(prob_out_i)


if __name__ == "__main__":
    main()



from abc import abstractmethod, ABCMeta


def abstract_metadata_key(*names):
    """Class decorator to add one or more abstract attribute.
    ref: https://stackoverflow.com/questions/45248243/most-pythonic-way-to-declare-an-abstract-class-property
    """

    def _func(cls, *names):
        """ Function that extends the __init_subclass__ method of a class."""
        cls.__abstract_metadata_keys__ = names
        for name in names:
            setattr(cls, name, NotImplemented)
        orig_init_subclass = cls.__init_subclass__

        def new_init_subclass(cls, **kwargs):
            try:
                orig_init_subclass(cls, **kwargs)
            except TypeError:
                orig_init_subclass(**kwargs)
            if getattr(cls, "metadata", NotImplemented) is NotImplemented:
                raise NotImplementedError(
                    "please define the metadata as a dictionary field in problem class"
                )
            for name in names:
                # if getattr(cls, name, NotImplemented) is NotImplemented:
                if name not in cls.metadata:
                    raise NotImplementedError(
                        f"{name} is required as a metadata entry but you haven't defined it"
                    )
        cls.__init_subclass__ = classmethod(new_init_subclass)
        return cls

    return lambda cls: _func(cls, *names)


@abstract_metadata_key(
    "problem_title",
    "problem_short_description",
    "author_names",
    "contact_name",
    "contact_email",
    "citations",
    "linked_sites",
)
class EspressoProblem(metaclass=ABCMeta):
    r"""Base class for all Espresso problems.
    All Espresso problems shoud be a subclass of this class.

    Parameters
    ----------
    example_number : int, optional
        The index of example you want to access. A typical Espresso problem will have
        several examples that have indices starting from 1. By default 1.

    Raises
    ------
    InvalidExampleError
        when you've passed in an example number that isn't included in current problem


    .. rubric:: Metadata

    Problem-sepecific metadata include the following keys:

    - ``problem_title``
    - ``problem_short_description``
    - ``author_names``
    - ``contact_name``
    - ``contact_email``
    - ``citations``
    - ``linked_sites``

    And they can be accessed through the :code:`metadata` dictionary:

    .. code-block:: pycon

       >>> from espresso import <ProblemClass>
       >>> <ProblemClass>.metadata["problem_title"]
       This is a problem about...
       >>> <ProblemClass>.metadata.keys()
       dict_keys(['problem_title', 'problem_short_description', 'author_names', 'contact_name', 'contact_email', 'citations', 'linked_sites'])

    .. rubric:: Required attributes
    
    Required methods and properties are guaranteed to be written by problem 
    contributors and available for user to access.

    .. autosummary::
        EspressoProblem.model_size
        EspressoProblem.data_size
        EspressoProblem.good_model
        EspressoProblem.starting_model
        EspressoProblem.data
        EspressoProblem.forward

    .. rubric:: Optional attributes
    
    Optional methods and properties have standards but are not always implemented for
    each Espresso problem. Try using them or check the documentation page for each
    problem to figure out whether they are available.

    .. autosummary::
        EspressoProblem.description
        EspressoProblem.covariance_matrix
        EspressoProblem.inverse_covariance_matrix
        EspressoProblem.jacobian
        EspressoProblem.plot_model
        EspressoProblem.plot_data
        EspressoProblem.misfit
        EspressoProblem.log_likelihood
        EspressoProblem.log_prior

    """

    def __init__(self, example_number=1):
        self.example_number = example_number
        self.params = dict()

    @property
    def description(self):
        """Returns a brief description of current example

        Returns
        -------
        str
            A string containing a brief (1-3 sentence) description of the example.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def model_size(self):
        """Returns the number of model parameters

        Returns
        -------
        int
            The number of model parameters (i.e. the dimension of a model vector).
        """
        raise NotImplementedError
    
    @property
    @abstractmethod
    def data_size(self):
        """Returns the number of data points

        Returns
        -------
        int
            The number of data points (i.e. the dimension of a data vector).
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def good_model(self):
        """Returns a model vector that the contributor regards as a sensible 
        explanation of the dataset

        Returns
        -------
        numpy.ndarray
            A model vector that the contributor regards as being a 'correct' or 
            'sensible' explanation of the dataset. (In some problems it may be the 
            case that there are many 'equally good' models. The contributor should 
            select just one of these.) It has the same shape as :attr:`model_size`.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def starting_model(self):
        """Returns a model vector representing a typical starting point for inversion

        Returns
        -------
        numpy.ndarray
            A model vector, possibly just np.zeros(model_size), representing a typical 
            starting point or 'null model' for an inversion. It has the same shape as 
            :attr:`model_size`.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def data(self):
        """Returns a data vector in the same format as output by :meth:`forward`

        Returns
        -------
        numpy.ndarray
            A data vector in the same shape as :attr:`data_size` and the output from
            :meth:`forward`
        """
        raise NotImplementedError

    @property
    def covariance_matrix(self):
        """Returns the covariance matrix for the data

        Returns
        -------
        numpy.ndarray
            The covariance matrix describing any uncertainty and correlations in the
            data vector. The output has shape (:attr:`data_size`, :attr:`data_size`)
        """
        raise NotImplementedError

    @property
    def inverse_covariance_matrix(self):
        """Returns the inverse data covariance matrix for the data

        Returns
        -------
        numpy.ndarray
            The inverse data covariance matrix, in the shape 
            (:attr:`data_size`, :attr:`data_size`)
        """
        raise NotImplementedError

    @abstractmethod
    def forward(self, model, with_jacobian=False):
        """Perform forward simulation with a model to produce synthetic data

        If with_jacobian == True, returns (d, G); else, returns d, where:

        - d : numpy.ndarray, shape(:attr:`data_size`,), a simulated data vector 
          corresponding to the given model
        - G : numpy.ndarray, shape(:attr:`data_size`, :attr:`model_size`), the
          Jacobian such that :math:`G[i,j] = \partial d[i]/\partial model[j]`
        
        If an example does not permit calculation of the Jacobian then calling with 
        with_jacobian=True should result in a NotImplementedError being raised.

        Parameters
        ----------
        model : numpy.ndarray
            a model vector, in the same shape of :attr:`model_size`
        with_jacobian: bool
            a switch governing the output required

        Returns
        -------
        (numpy.ndarray, numpy.ndarray) | numpy.ndarray
            (d, G) or d, depending on the value of with_jacobian. Details above.
        """
        raise NotImplementedError

    def jacobian(self, model):
        """Returns the Jacobian matrix

        Parameters
        ----------
        model : numpy.ndarray
            a model vector, in the same shape of :attr:`model_size`

        Returns
        -------
        numpy.ndarray
            the Jacobian such that :math:`G[i,j] = \partial d[i]/\partial model[j]`, in
            the shape of (:attr:`data_size`, :attr:`model_size`)
        """
        raise NotImplementedError

    def plot_model(self, model):
        """Returns a figure containing a basic visualisation of the model

        Parameters
        ----------
        model : numpy.ndarray
            a model vector for visualisatioin, in the same shape of :attr:`model_size`

        Returns
        -------
        matplotlib.figure.Figure
            A figure handle containing a basic visualisation of the model.
        """
        raise NotImplementedError
    
    def plot_data(self, data, data2 = None):
        """Returns a figure containing a basic visualisation of a dataset and 
        (optionally) comparing it to a second dataset

        Parameters
        ----------
        data : numpy.ndarray
            A data vector for visualisation
        data2 : numpy.ndarray, optional
            A second data vector, for comparison with the first, by default None
        
        Returns
        -------
        matplotlib.figure.Figure
            A figure handle containing a basic visualisation of a dataset and 
            (optionally) comparing it to a second dataset.
        """
        raise NotImplementedError

    def misfit(self, data, pred):
        """Returns a measure of the extent to which a predicted data vector agrees with
        observed data

        Parameters
        ----------
        data : numpy.ndarray
            An observed data vector to base on
        pred : numpy.ndarray
            A predicted data vector to evaluate
        
        Returns
        -------
        Number
            A measure of the extent to which a predicted data vector, ``pred``, agrees 
            with observed data, ``data``. Smaller numbers imply better agreement; 
            0 -> perfect match.
        """
        raise NotImplementedError
    
    def log_likelihood(self, data, pred):
        """Returns the log likelihood density value

        Parameters
        ----------
        data : numpy.ndarray
            An observed data vector to base on
        pred : numpy.ndarray
            A predicted data vector to evaluate
        
        Returns
        -------
        Number
            The log likelihood that ``data`` is an imperfect observation of a system 
            generating data ``pred``.
        """
        raise NotImplementedError

    def log_prior(self, model):
        """Returns the log prior density value

        Parameters
        ----------
        model : numpy.ndarray
            A model vector to evaluate, shape(:attr:`model_size`)

        Returns
        -------
        Number
            The log probability that a system is described by ``model`` prior to 
            seeing any data.
        """
        raise NotImplementedError
        

    def __getattr__(self, key):
        if hasattr(self, "params") and key in self.params:
            return self.params[key]
        if key in self.metadata:
            return self.metadata[key]
        else:
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute '{key}'"
            )



r"""
Module to contain funtions that help generate documentation
"""
from pathlib import Path
from shutil import copy
import os
import espresso as esp


def read_metadata(contrib_name, lines):
    # - problem_title
    # - problem_short_description
    # - author_names
    # - contact_name
    # - contact_email
    # - [optional] citations -> []
    # - [optional] linked_sites -> [(name, link)]
    contrib_class_name = contrib_name.title().replace("_", "")
    contrib_class = getattr(esp, contrib_class_name)
    class_metadata = contrib_class.metadata
    lines.append(":::{admonition} Contribution Metadata for ")
    lines[-1] += "*" + class_metadata["problem_title"] + "* \n:class: important"
    # metadata - short description
    lines.append(class_metadata["problem_short_description"])
    lines.append("```{eval-rst}")
    # metadata - authors
    lines.append("\n:Author: " + ", ".join(class_metadata["author_names"]))
    # metadata - contact
    lines.append(
        ":Contact: "
        + class_metadata["contact_name"]
        + " ("
        + class_metadata["contact_email"]
        + ")"
    )
    # metadata - citations
    if len(class_metadata["citations"]) > 0:
        lines.append(":Citation:")
        if len(class_metadata["citations"]) == 1:
            citation = class_metadata["citations"][0]
            lines.append(f"  {citation[0]}")
            if citation[1]:
                lines[-1] += f", {citation[1]}"
        else:
            for citation in class_metadata["citations"]:
                lines.append(f"  - {citation[0]}")
                if citation[1]:
                    lines[-1] += f", {citation[1]}"
    # metadata - extra website
    if len(class_metadata["linked_sites"]) > 0:
        lines.append(":Extra website:")
        if len(class_metadata["linked_sites"]) == 1:
            name, link = class_metadata["linked_sites"][0]
            lines.append(f"  `{name} <{link}>`_")
        else:
            for (name, link) in class_metadata["linked_sites"]:
                lines.append(f"  - [{name}]({link})")
    # ending
    lines.append("```")
    lines.append(":::")


def read_file(contrib_dir, dest_contrib_dir, file_name, lines):
    src_path = contrib_dir / file_name
    dst_path = dest_contrib_dir / file_name
    copy(src_path, dst_path)
    lines.append("```{include} ./" + file_name + "\n```")


def contribs(BASE_PATH, DEST_PATH):
    names = [cls.__module__.split(".")[-1] for cls in esp.list_problems()]
    all_contribs = [
        (contrib, Path(f"{BASE_PATH}/{contrib}"), Path(f"{DEST_PATH}/{contrib}"))
        for contrib in names
    ]
    all_contribs = [
        contrib
        for contrib in all_contribs
        if contrib[1].exists() and contrib[1].is_dir()
    ]
    return all_contribs


def gen_contrib_docs(BASE_PATH, DEST_PATH):
    os.mkdir(DEST_PATH)
    # prepare index file
    index_lines = []
    # move info for each contribution
    for (contrib, src_folder, dst_folder) in contribs(BASE_PATH, DEST_PATH):
        # make new folder docs/source/contrib/<contrib-name>
        os.mkdir(dst_folder)
        lines = []
        # Provide hint about where this file came from
        lines.append(
            "<!--- This file was automatically generated by \n"
            "      _ext/generate_contrib_docs.py based on the\n"
            "      contributor-supplied metadata and the files\n"
            "      README.md and LICENCE. \n"
            "-->\n\n"
        )
        # include README.md
        read_file(src_folder, dst_folder, "README.md", lines)
        # format metadata content
        read_metadata(contrib, lines)
        # include LICENCE
        lines.append("\n## LICENCE\n")
        read_file(src_folder, dst_folder, "LICENCE", lines)
        # write to index.md file
        with open(f"{dst_folder}/index.md", "w") as f:
            f.write("\n".join(lines))
        # add contrib link to contrib/index.rst
        index_lines.append(f"    generated/{contrib}/index.md\n")
    # write index file
    with open(Path(DEST_PATH).parent / "_index.rst", "r") as f:
        index_template = f.read()
    with open(Path(DEST_PATH).parent / "index.rst", "w") as f:
        f.write(
            "..\n   *** N.B. This file is automatically generated by\n"
            "       _ext/generate_contrib_docs.py based on the template\n"
            "       in _index.rst ***\n\n"
        )
        f.write(index_template)
        for line in index_lines:
            f.write(line)



_all_problems = []
_capability_matrix = dict()

def list_problem_names(capabilities: list = None):
    """Returns a list of all Espresso problem names"""
    _problems = list_problems(capabilities)
    _all_names = [p.__name__ for p in _problems]
    return _all_names

def list_problems(capabilities: list = None):
    """Returns a list of all Espresso problem classes"""
    if capabilities is None:
        return _all_problems
    elif not isinstance(capabilities, (list, set, tuple)):
        raise ValueError(
            "pass in a list of capabilities, e.g. "
            "`espresso.list_problems(['plot_model'])"
        )
    else:
        _problem_names = []
        for p, c in _capability_matrix.items():
            ok = True
            for to_check in capabilities:
                if not (to_check in c and c[to_check] == 1):
                    ok = False
                    break
            if ok:
                _problem_names.append(p)
        _problems = []
        for p in _all_problems:
            if p.__name__ in _problem_names:
                _problems.append(p)
        return _problems

def list_capability(problem_names: list) -> dict:
    return {k:v for k,v in _capability_matrix.items() if k in problem_names}

# from .example_name import ExampleName

# _all_problems = [ ExampleName, ]

# _capability_matrix = {
#     "ExampleName": {
#         "model_size": 1,
#         "data_size": 1,
#         "starting_model": 1,
#         ...
#     },
#     ...
# }

from ._magnetotelluric_1D import Magnetotelluric1D
from ._slug_test import SlugTest
from ._fm_wavefront_tracker import FmWavefrontTracker
from ._gravity_density import GravityDensity
from ._pumping_test import PumpingTest
from ._xray_tracer import XrayTracer
from ._receiver_function import ReceiverFunction
from ._simple_regression import SimpleRegression

_all_problems = [
    Magnetotelluric1D,
    SlugTest,
    FmWavefrontTracker,
    GravityDensity,
    PumpingTest,
    XrayTracer,
    ReceiverFunction,
    SimpleRegression,
]

_capability_matrix = {
    "Magnetotelluric1D": {
        "model_size": 1,
        "data_size": 1,
        "good_model": 1,
        "starting_model": 1,
        "data": 1,
        "forward": 1,
        "description": 1,
        "covariance_matrix": 1,
        "inverse_covariance_matrix": 1,
        "jacobian": 1,
        "plot_model": 1,
        "plot_data": 1,
        "misfit": 1,
        "log_likelihood": 0,
        "log_prior": 0,
        "set_start_model": 1,
        "set_start_mesh": 1,
        "set_obs_data": 1
    },
    "SlugTest": {
        "model_size": 1,
        "data_size": 1,
        "good_model": 1,
        "starting_model": 1,
        "data": 1,
        "forward": 0,
        "description": 1,
        "covariance_matrix": 1,
        "inverse_covariance_matrix": 1,
        "jacobian": 0,
        "plot_model": 1,
        "plot_data": 0,
        "misfit": 0,
        "log_likelihood": 1,
        "log_prior": 0
    },
    "FmWavefrontTracker": {
        "model_size": 1,
        "data_size": 1,
        "good_model": 1,
        "starting_model": 1,
        "data": 1,
        "forward": 1,
        "description": 0,
        "covariance_matrix": 0,
        "inverse_covariance_matrix": 0,
        "jacobian": 1,
        "plot_model": 1,
        "plot_data": 0,
        "misfit": 0,
        "log_likelihood": 0,
        "log_prior": 0,
        "clean_tmp_files": 1,
        "tmp_files": 1,
        "tmp_paths": 1,
        "exe_fm2dss": 1
    },
    "GravityDensity": {
        "model_size": 1,
        "data_size": 1,
        "good_model": 1,
        "starting_model": 1,
        "data": 1,
        "forward": 1,
        "description": 0,
        "covariance_matrix": 1,
        "inverse_covariance_matrix": 0,
        "jacobian": 1,
        "plot_model": 1,
        "plot_data": 1,
        "misfit": 0,
        "log_likelihood": 0,
        "log_prior": 0
    },
    "PumpingTest": {
        "model_size": 1,
        "data_size": 1,
        "good_model": 1,
        "starting_model": 1,
        "data": 1,
        "forward": 0,
        "description": 1,
        "covariance_matrix": 1,
        "inverse_covariance_matrix": 1,
        "jacobian": 0,
        "plot_model": 1,
        "plot_data": 0,
        "misfit": 0,
        "log_likelihood": 1,
        "log_prior": 0
    },
    "XrayTracer": {
        "model_size": 1,
        "data_size": 1,
        "good_model": 1,
        "starting_model": 1,
        "data": 1,
        "forward": 0,
        "description": 1,
        "covariance_matrix": 0,
        "inverse_covariance_matrix": 0,
        "jacobian": 0,
        "plot_model": 1,
        "plot_data": 0,
        "misfit": 0,
        "log_likelihood": 0,
        "log_prior": 0
    },
    "ReceiverFunction": {
        "model_size": 1,
        "data_size": 1,
        "good_model": 1,
        "starting_model": 1,
        "data": 1,
        "forward": 0,
        "description": 1,
        "covariance_matrix": 1,
        "inverse_covariance_matrix": 0,
        "jacobian": 0,
        "plot_model": 1,
        "plot_data": 1,
        "misfit": 0,
        "log_likelihood": 1,
        "log_prior": 1,
        "rf": 1
    },
    "SimpleRegression": {
        "model_size": 1,
        "data_size": 1,
        "good_model": 1,
        "starting_model": 1,
        "data": 1,
        "forward": 1,
        "description": 1,
        "covariance_matrix": 1,
        "inverse_covariance_matrix": 1,
        "jacobian": 1,
        "plot_model": 0,
        "plot_data": 0,
        "misfit": 0,
        "log_likelihood": 0,
        "log_prior": 0
    }
}


"""Check if all requirements are listed in pyproject.toml

This script assumes you have geo-espresso installed via:
$ python espresso_machine/build_package/build.py

Usage: python check_requires.py [-c <example_name] [--file <file_name>]
"""

import sys
import pathlib
import pytest
from stdlib_list import stdlib_list

import _utils

known_dependencies = {
    "numpy",
    "scipy",
    "scipy.stats",
    "scipy.sparse",
    "scipy.interpolate",
    "scipy.constants",
    "matplotlib",
    "matplotlib.pyplot",
    "tqdm",
    "mpmath",
}
to_exclude = {"espresso", "run_examples"}


def _strip_pkg(modules):
    res = set()
    for mod in modules:
        pkg = mod.split(".")[0]
        if not pkg.startswith("_"):
            res.add(pkg)
    return res


def _get_inbuilt_pkg():
    return _strip_pkg(stdlib_list("3.7"))


def _get_known_depended_pkg():
    for pkg in known_dependencies:
        __import__(pkg)
    return _strip_pkg(set(sys.modules.keys()))


def _get_imported_pkg(problem_specified):
    import run_examples

    run_examples.main([problem_specified], _utils.DEFAULT_TIMEOUT)
    return _strip_pkg(set(sys.modules.keys()))


def _get_requirements(problem_specified):
    inbuilt = _get_inbuilt_pkg()
    known_depended = _get_known_depended_pkg()
    all_imported = _get_imported_pkg(problem_specified)
    return inbuilt, known_depended, all_imported


def get_extra_requirements(problem_specified):
    inbuilt, known_depended, all_imported = _get_requirements(problem_specified)
    new_dependencies = all_imported - known_depended - inbuilt
    not_listed = new_dependencies - to_exclude
    return not_listed


def test_requires(contrib):
    not_listed = get_extra_requirements(contrib[0])
    assert len(not_listed) == 0, f"new dependency to be listed: {not_listed}"
    print("√ Passed requirements test.")


if __name__ == "__main__":
    pytest.main([pathlib.Path(__file__)])



class EspressoError(Exception):
    """Base class for all Espresso errors
    """
    pass

# Multiple inheritance means this can be caught by all of the following:
# ... except InvalidExampleError
# ... except EspressoError
# ... except ValueError
class InvalidExampleError(EspressoError, ValueError):
    r"""Raised if user attempts to instantiate an example number that does not exist

    This is a subclass of both :exc:`EspressoError` and :exc:`ValueError`.

    Examples
    --------

    >>> from espresso import SimpleRegression, InvalidExampleError
    >>> try:
    ...     reg = SimpleRegression(6)
    ... except InvalidExampleError:
    ...     print("InvalidExampleError triggered")
    ... 
    InvalidExampleError triggered
    
    """
    def __init__(self, *args):
        super().__init__(*args)
    def __str__(self):
        super_msg = super().__str__()
        msg = "Unrecognised example number.\n\nPlease refer to the Espresso documentation " \
              "(https://geo-espresso.readthedocs.io/)\nfor full details of the examples " \
              "provided within this test problem."
        if len(super_msg)>0:
            return msg+"\n\n"+super_msg
        else:
            return msg





# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import datetime
import os
import sys
import subprocess
from pathlib import Path

import espresso as esp


# -- Generate API references doc ---------------------------------------------
def run_autogen(_):
    cmd_path = "sphinx-autogen"
    if hasattr(sys, "real_prefix"):  # Check to see if we are in a virtualenv
        # If we are, assemble the path manually
        cmd_path = os.path.abspath(os.path.join(sys.prefix, "bin", cmd_path))
    subprocess.check_call(
        [
            cmd_path, "-i", 
            "-t", Path(__file__).parent / "_templates", 
            "-o", Path(__file__).parent / "user_guide" / "api" / "generated", 
            Path(__file__).parent / "user_guide" / "api" / "index.rst"
        ]
    )

def setup(app):
    # This function is automatically called by sphinx-build at the start of the
    # build process.
    app.connect("builder-inited", run_autogen)


# -- Project information -----------------------------------------------------
project = 'Espresso'
copyright = f"{datetime.date.today().year}, InLab, {project} development team"
version = "dev" if "dev" in esp.__version__ else f"v{esp.__version__}"


# -- General configuration ---------------------------------------------------
sys.path.append(os.path.abspath("./_ext"))
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx.ext.doctest",
    "sphinx.ext.mathjax",
    "sphinx_design",
    "sphinx_togglebutton",
    "sphinx_copybutton",
    "sphinx.ext.napoleon",
    "myst_nb",
    # "sphinx_gallery.gen_gallery",
    "sphinxcontrib.mermaid",
    "generate_contrib_docs",                # our own extension
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    ".DS_Store",
    "user_guide/contrib/_index.rst",
    "contributor_guide/contribute.rst",     # TODO migrate to new_contrib.rst
]

source_suffix = ".rst"
source_encoding = "utf-8"
master_doc = "index"
pygments_style = "trac"        # https://pygments.org/styles/
add_function_parentheses = False

# Configuration to include links to other project docs
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "matplotlib": ("https://matplotlib.org/stable", None),
}

# settings for the sphinx-copybutton extension
copybutton_prompt_text = ">>> "


# -- Options for HTML output -------------------------------------------------
html_title = f'{project} <span class="project-version">{version}</span>'
html_short_title = project
html_favicon = "_static/inlab_logo_60px.png"

html_theme = "sphinx_book_theme"
html_theme_options = {
    "repository_url": "https://github.com/inlab-geo/espresso",
    "repository_branch": "main",
    "path_to_docs": "docs/source/",
    "launch_buttons": {
        "notebook_interface": "classic",
        "inlab_url": "http://www.inlab.edu.au/",
    },
    "extra_footer": "",
    "home_page_in_toc": True,
    "use_repository_button": True,
    "use_edit_page_button": True,
    "use_source_button": True,
    "use_issues_button": True,
    "use_download_button": True,
    "use_sidenotes": True,
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/inlab-geo/geo-espresso",
            "icon": "https://img.shields.io/badge/GitHub-espresso-171515?logo=github&labelColor=f8f9fa&style=flat-square&logoColor=171515",
            "type": "url",
        },
        {
            "name": "Version",
            "url": "https://pypi.org/project/geo-espresso/",
            "icon": "https://img.shields.io/pypi/v/geo-espresso?logo=pypi&style=flat-square&color=83C5BE&labelColor=f8f9fa&label=latest",
            "type": "url",
        },
    ],
}

html_static_path = ["_static"]
html_css_files = ["style.css"]
html_context = {
    "display_github": True, # Integrate GitHub
    "github_user": "inlab-geo", # Username
    "github_repo": "espresso", # Repo name
    "github_version": "main", # Version
    "conf_py_path": "/source/", # Path in the checkout to the docs root
}


# -- myst-nb settings ---------------------------------------------------------
myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "html_image",
]


# -- Cutomised variables ------------------------------------------------------
rst_epilog = """
.. _repository: https://github.com/inlab-geo/espresso
.. _newissue: https://github.com/inlab-geo/espresso/issues/new/choose
.. _Slack: https://join.slack.com/t/inlab-community/shared_invite/zt-1ejny069z-v5ZyvP2tDjBR42OAu~TkHg
"""



r"""Submodule for utility functions in espresso.

Some of these functions are used by contributors to conveniently implement their
problem class. Others can be utilised by users to perform analysis on inversion
results.

.. important::

    Please always use :func:`absolute_path` or :func:`loadtxt` when trying to load data
    from a relative path into an Espresso problem class.

"""

from .data_loader import loadtxt, absolute_path
from .file_handler import silent_remove

__all__ = [
    "loadtxt",
    "absolute_path",
    "silent_remove",
]



import matplotlib.pyplot as plt
import numpy as np

import cofi
from espresso import Magnetotelluric1D

# Load example
mt = Magnetotelluric1D(example_number=1)

# Plot true model and synthetic response
fig1 = mt.plot_model(mt.good_model, title='True model')
plt.show()

# define CoFI BaseProblem
mt_problem = cofi.BaseProblem()
mt_problem.set_initial_model(mt.starting_model)

# define regularization: smoothing
smoothing_factor = 10
reg_smoothing = cofi.utils.QuadraticReg(smoothing_factor, model_size = mt.model_size, reg_type = "smoothing")
reg = reg_smoothing

# Define objective function
def objective_func(model, reg):
    dpred = mt.forward(model)
    data_misfit = mt.misfit(mt.data,dpred,mt.inverse_covariance_matrix)
    model_reg = reg(model)
    return  data_misfit + model_reg

mt_problem.set_objective(objective_func, args=[reg])

# Define the inversion options
my_options = cofi.InversionOptions()
my_options.set_tool("scipy.optimize.minimize")
my_options.set_params(method="L-BFGS-B",options={'ftol':1e-3,'maxiter': 100})

# Run the inversion
print("Running inversion...")
inv = cofi.Inversion(mt_problem, my_options)
inv_result = inv.run()
print("   done!")


# Plot the results
fig2 = mt.plot_model(inv_result.model, title='Inversion model')
fig3 = mt.plot_data(mt.data, mt.forward(inv_result.model), Cm = mt.covariance_matrix)
nRMSE = np.sqrt(mt.misfit(mt.data, mt.forward(inv_result.model), Cm_inv = mt.inverse_covariance_matrix)/mt.data_size)
print('   nRMSE = %.3f'%nRMSE)

plt.show()



"""Validate a few / all contributions by 
1. Running the problem(s) and checking outputs
   - according to criteria.py
   - tested by test_examples.py
2. If post-build, checking all required packages are specified
   - tested by check_requires.py

Usage: 
python validate.py [-h] [--all] [--pre] [--post] [--contrib CONTRIBS] [--file <file_name>]
"""

import sys
import subprocess
import pathlib
import pytest

import _utils


def prep_params(pre_build=None):
    pre = _utils.pre_build() if pre_build is None else pre_build
    dir_parent = pathlib.Path(__file__).parent.resolve()
    specified_contribs = _utils.problems_to_run_names_only()
    extra_args = []
    if specified_contribs is not None:
        for contrib in specified_contribs:
            extra_args.append("--contribution")
            extra_args.append(contrib)
    return pre, dir_parent, extra_args


def test_all_examples(pre, dir_parent, extra_args):
    py_test_examples = dir_parent / "test_examples.py"
    pytest_cmd = [sys.executable, "-m", "pytest", str(py_test_examples)]
    pytest_cmd.extend(extra_args)
    if not pre:
        pytest_cmd.append("--post")
    exit_status_test_examples = subprocess.run(pytest_cmd).returncode
    if exit_status_test_examples != pytest.ExitCode.OK:
        sys.exit(exit_status_test_examples)


def test_requirements(pre, dir_parent, extra_args):
    if not pre:
        py_check_requires = dir_parent / "check_requires.py"
        pytest_cmd = [sys.executable, "-m", "pytest", str(py_check_requires)]
        pytest_cmd.extend(extra_args)
        exit_status_check_requires = subprocess.run(pytest_cmd).returncode
        if exit_status_check_requires != pytest.ExitCode.OK:
            sys.exit(exit_status_check_requires)


# --> main test
def main(pre_build=None):
    pre, dir_parent, extra_args = prep_params(pre_build)
    test_all_examples(pre, dir_parent, extra_args)
    test_requirements(pre, dir_parent, extra_args)


if __name__ == "__main__":
    main()



from espresso import EspressoProblem
from espresso.exceptions import InvalidExampleError

import numpy as np
import matplotlib.pyplot as plt
from scipy.constants import mu_0


class Magnetotelluric1D(EspressoProblem):
    """Forward simulation class
    """

    metadata = {
        "problem_title": "1D magnetotelluric",               
        "problem_short_description": "Compute the MT response of a 1D resistivty model",    

        "author_names": ["Hoël Seillé"],    

        "contact_name": "Hoël Seillé",         
        "contact_email": "hoel.seille@csiro.au",

        "citations": [],

        "linked_sites": [], 
    }


    def __init__(self, example_number=1):
        super().__init__(example_number)


        if example_number == 1:
            
            # true model
            # layer electrical resitivity in log10 ohm.m to ensures positive resistivities
            nLayers, min_thickness, vertical_growth= 50, 5, 1.15
            thicknesses = [min_thickness * vertical_growth**i for i in range(nLayers-1)]
            true_depths = np.cumsum(thicknesses)
            true_model = np.ones((nLayers))*3
            true_model[:10] = 2
            true_model[10:17] = 1
            
            # define frequencies (in Hz) where responses are computed
            fmin, fmax, f_per_decade = 0.1, 1e4, 5
            freqs = get_frequencies(fmin,fmax,f_per_decade)
            
            # generate synthetic data 
            # calculate impedance Z
            Z = forward_1D_MT(true_model, true_depths, freqs, return_Z = True)
            # add noise
            Z, Zerr = add_noise(Z, percentage = 5, seed = 1234)
            # transform Z to log10 apparent resistivity and phase (dobs)
            dobs, derr = z2rhophy(freqs, Z, dZ=Zerr)

            # define a starting 1D mesh and model 
            # the mesh contains many cells to produce a smooth model
            nLayers, min_thickness, vertical_growth= 50, 5, 1.15
            thicknesses = [min_thickness * vertical_growth**i for i in range(nLayers-1)]
            starting_depths = true_depths
            starting_model = np.ones((nLayers)) * 2 # 100 ohm.m starting model (log10 scale) 

            self._desc = "1 MT sounding over a 3 layers Earth model, smooth inversion"
            self._mtrue = true_model
            self._dptrue = true_depths
            self._mstart = starting_model
            self._dpstart = starting_depths
            self._dobs = dobs
            self._derr = derr
            self._freqs = freqs


        else:
            raise InvalidExampleError

    @property
    def description(self):
        return self._desc

    @property
    def model_size(self):
        return len(self.good_model)

    @property
    def data_size(self):
        return len(self.data)

    @property
    def good_model(self):
        return self._mtrue

    @property
    def starting_model(self):
        return self._mstart.flatten()
    
    @property
    def data(self):
        return self._dobs

    @property
    def covariance_matrix(self):
        return np.identity(len(self.data)) * self._derr**2

    @property
    def inverse_covariance_matrix(self):
        return np.identity(len(self.data)) * (1/self._derr**2)
        
    def set_start_model(self, new_start_model):
        self._mstart = new_start_model
        self._mtrue = new_start_model

    def set_start_mesh(self, new_start_mesh):
        self._dpstart = new_start_mesh
        self._dptrue = new_start_mesh

    def set_obs_data(self, dobs, derr, freqs):
        self._dobs = dobs
        self._derr = derr
        self._freqs = freqs

    def forward(self, model, with_jacobian = False):
        if with_jacobian:
            dpred, G = forward_1D_MT(model, self._dptrue, self._freqs, return_G=True)
            return dpred, G
        else:
            dpred= forward_1D_MT(model, self._dptrue, self._freqs, return_G=False)
            return dpred
    
    def jacobian(self, model):
        dpred, G = forward_1D_MT(model, self._dptrue, self._freqs, return_G = True)
        return G

    def plot_model(self, model, depths = None, max_depth = -1000, res_bounds = [0,4], title = None):
        nLayers = len(model) 
        if depths is None:
            max_z = min(max_depth,-self._dptrue[-1])
            depths = np.r_[-self._dptrue,max_z]
        else:
            max_z = min(max_depth,-depths[-1])
            depths = np.r_[-depths,max_z]

        fig = plt.figure(1,figsize=(4,4))
        ax = fig.add_subplot(1, 1, 1)
        px = np.zeros([2*nLayers,2])
        px[0::2,0],px[1::2,0],px[1::2,1],px[2::2,1] = model,model,depths[:],depths[:-1]
        ax.plot((px[:,0]),px[:,1],'k-',lw=3)
        ax.set_xlim(res_bounds);
        ax.set_ylim(max_depth,0)
        ax.set_xlabel(r'Log$_{10}$ resistivity ($\Omega$m)')
        ax.set_ylabel('Depth (m)')
        ax.grid(lw=0.2)
        ax.set_title(title)
        plt.tight_layout()
        #plt.show()
        return fig
    
    def plot_data(self, data, data2 = None, Cm = None):
        nf = len(self._freqs)
        log10_rho = data[:nf]
        phase = data[nf:]
        fig, axs = plt.subplots(2,1,sharex = True,figsize = (4,4), num = 2)
        if Cm is None:
            axs[0].plot(np.log10(1/self._freqs), log10_rho,'k.')
            axs[1].plot(np.log10(1/self._freqs), phase,'k.')
        else:
            d_log10_rho = (np.diag(Cm)**0.5)[:nf]
            d_phase = (np.diag(Cm)**0.5)[nf:]
            axs[0].errorbar(np.log10(1/self._freqs), log10_rho, yerr = d_log10_rho, 
                            fmt='wo',elinewidth=0.6,markersize=4 ,ecolor='k',
                            capsize=2,capthick=0.6,mec='k',mew=0.5, alpha=0.9, zorder = 1, label = 'Obs')
            axs[1].errorbar(np.log10(1/self._freqs), phase, yerr = d_phase, 
                            fmt='wo',elinewidth=0.6,markersize=4 ,ecolor='k',
                            capsize=2,capthick=0.6,mec='k',mew=0.5, alpha=0.9, zorder = 1)
        if data2 is not None:
            log10_rho2 = data2[:nf]
            phase2 = data2[nf:]
            axs[0].plot(np.log10(1/self._freqs), log10_rho2, 'k-', lw = 1, zorder = 2, label = 'Resp')
            axs[1].plot(np.log10(1/self._freqs), phase2, 'k-', lw = 1, zorder = 2)
        axs[0].set_ylabel(r'Log$_{10}$ $\rho_{app}$ ($\Omega$m)')
        axs[1].set_yticks([0,45,90])
        axs[1].set_ylabel('Phase (deg.)')
        axs[1].set_xlabel('Log$_{10}$ Period (s)')
        axs[0].grid(lw=0.2)
        axs[1].grid(lw=0.2)
        # axs[0].legend()
        plt.tight_layout()
        #plt.show()
        return fig

    def misfit(self, data, data2, Cm_inv = None):
        res = data - data2
        if Cm_inv is None:
            return float(res.T @ res)
        else:
            return float(res.T @ Cm_inv @ res)          

    def log_likelihood(self,data,data2):
        raise NotImplementedError               # optional
    
    def log_prior(self, model):
        raise NotImplementedError               # optional



def get_frequencies(fmin, fmax, f_per_decade):
    nf = ((np.log10(fmax) - np.log10(fmin)) * f_per_decade) + 1
    freqs = np.logspace(np.log10(fmax), np.log10(fmin), num = int(nf))
    return freqs


def forward_1D_MT(model, depths, freqs, return_G = False, return_Z = False):
    # Impedance recursive approach to compute analytically MT transfer
    # functions over a layered 1D Earth, as in:
    # Pedersen, J., Hermance, J.F. Least squares inversion of one-dimensional 
    #       magnetotelluric data: An assessment of procedures employed 
    #       by Brown University. Surv Geophys 8, 187–231 (1986).

    # adapted from https://empymod.emsig.xyz/en/stable/gallery/fdomain/magnetotelluric.html
    
    w = 2*np.pi*freqs # angular frequencies
    model_lin = 10**model # electrical resistivities in linear scale

    # Calculate impedance Z at the top of the bottom half space 
    Z = np.sqrt(1j * w * mu_0 * model_lin[-1])
    
    # The surface impedance Z is found recursively from the bottom propagating upwards
    for j in range(len(depths)-1,-1,-1):
        # calculate thickness th of layer j
        if j == 0: th = depths[j]
        else: th = depths[j] - depths[j-1]
        # calculate intrinsic impedance zo of layer j
        zo = np.sqrt(1j * w * mu_0 * model_lin[j])
        # calculate reflection coefficient R of layer j
        R = (zo - Z) / (zo + Z)
        # calculate induction parameter gamma of layer j
        gamma = np.sqrt(1j * w * mu_0 / model_lin[j])
        # update impedance Z at the top of layer j
        Z = zo * (1 - R * np.exp(-2*gamma*th)) / (1 + R * np.exp(-2*gamma*th))
        
    # instead of using the complex impedance, we use the log amplitude and phase
    # of Z (log10 apparent resistivity and phase)(Wheelock, Constable, Key; GJI 2015)
    data, _ = z2rhophy(freqs, Z, dZ=None)
    
    if return_G:
        # Jacobian calculation using finite differences 
        M = len(model)
        N = len(data)
        G = np.zeros((N, M))
        
        # ppert = 0.01 # Perturbation (percentage)
        # for i in range(M):
        #     model_pert = model + model * ppert * np.identity(M)[:,i]
        #     dpert = forward_1D_MT(model_pert,depths, freqs)
        #     G[:,i] = (dpert - data) / (model * ppert)[i]

        apert = 0.01 # Perturbation (absolute)
        for i in range(M):
            model_pert = model + apert * np.identity(M)[:,i]
            dpert = forward_1D_MT(model_pert,depths, freqs)
            G[:,i] = (dpert - data) / apert

    if return_Z:
        return Z

    elif return_G:
        return data, G

    else:
        return data



def z2rhophy(freqs,Z,dZ=None):
    
    # calcul of apparent resistivity and phases from the impedance Z
    rho_app = abs(Z)**2 / (mu_0*2*np.pi*freqs)
    phase = np.degrees(np.arctan2(Z.imag,Z.real))
    
    # calculate errors of apparent resistivity and phases
    if dZ is None:
        return np.r_[np.log10(rho_app), phase], np.r_[np.zeros(Z.shape[0]*2)]
    else:
        drho_app = 2*rho_app*dZ / abs(Z)
        dphase = np.degrees(0.5 * (drho_app/rho_app))
        log10_drho_app = (1/np.log(10)) * (drho_app/rho_app)
        return np.r_[np.log10(rho_app), phase], np.r_[log10_drho_app, dphase]




def add_noise(Z, percentage = 5, seed = 1234):
    # The standard deviations are taken as a percentage of the amplitude |Z|.
    # We assume a circularly symmetric Gaussian distribution in the complex 
    # plane for Z, with a common standard deviation Zerr for the real and
    # imaginary parts.
    
    np.random.seed(seed)
    Zerr = np.zeros_like(Z, dtype=float)
    for f in range(Z.shape[0]):
        Zerr[f] = 0.01*percentage*(Z[f].real**2+Z[f].imag**2)**0.5
        Zr = Z[f].real + np.random.normal(0.0, 0.01*percentage*abs(Z[f]))
        Zi = Z[f].imag + np.random.normal(0.0, 0.01*percentage*abs(Z[f]))
        Z[f] = Zr + 1j*Zi
    return Z, Zerr


def load_data(filename, error_floor = 0.05, subsampling = 3):

    # load the data, transform the apparent resistivity and its error to log10
    MTdata = np.loadtxt('../data/%s'%(filename),skiprows=1)
    MTdata = MTdata[::subsampling,:]
    
    freqs = MTdata[:,0]
    rho = MTdata[:,1]
    rho_err = MTdata[:,2]
    phase = MTdata[:,3]
    phase_err = MTdata[:,4]
    
    dobs = np.r_[np.log10(rho), phase]
    
    log10_drho = (1/np.log(10)) * (rho_err/rho)
    
    # add an error floor to the data: if the error if lower than the floor, then it is raised to the floor
    # it prevents from including unrealistic small data error into the inversion
    error_floor_rho = error_floor # the error floor is set as a percentage of the apparent resistivity 
    ef_rho_log = np.log10(1 + error_floor_rho) 
    ef_phase = (100 * error_floor_rho) * 0.286    
    
    err_rho_log = np.maximum(log10_drho, ef_rho_log)
    err_phase = np.maximum(phase_err, ef_phase)
    
    derr = np.r_[err_rho_log, err_phase]
    
    return freqs, dobs, derr






from pathlib import Path
import importlib
import sys

import espresso


# APV removed following as it is never used(?)
# CONTRIBS = [nm for nm in dir(esp) if not nm.startswith("_") and nm[0].islower()]
BASE_PATH = espresso.__path__[0]
DEST_PATH = Path(__file__).resolve().parent.parent / "user_guide" / "contrib" / "generated"
ROOT_PATH = Path(__file__).resolve().parent.parent.parent.parent
_machine_doc_utils_gen_docs = ROOT_PATH / "espresso_machine" / "doc_utils"

try:
    sys.path.append(str(_machine_doc_utils_gen_docs))
    import gen_docs
except:
    from espresso._machine.doc_utils import gen_docs


def setup(app):
    app.connect("builder-inited", lambda _: gen_docs.gen_contrib_docs(BASE_PATH, DEST_PATH))
    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



import numpy as np
from scipy.stats import multivariate_normal

from espresso import EspressoProblem
from espresso.exceptions import InvalidExampleError
from espresso.utils import absolute_path as path, silent_remove
from . import waveTracker as wt


class FmWavefrontTracker(EspressoProblem):
    """Forward simulation class
    """

    metadata = {
        "problem_title": "Fast Marching Wave Front Tracking",                # To be used in docs
        "problem_short_description": "The wave front tracker routines solves boundary"\
            " value ray tracing problems into 2D heterogeneous wavespeed media, "\
            "defined by continuously varying velocity model calculated by 2D cubic "\
            "B-splines.",    # 1-3 sentences

        "author_names": ["Nick Rawlinson","Malcolm Sambridge"],    # List of names e.g. author_names = ["Sally Smith", "Mark Brown"]

        "contact_name": "Malcolm Sambridge",         # Contact for contributor/maintainer of espresso example
        "contact_email": "Malcolm.Sambridge@anu.edu.au",

        "citations": [
            (
                "Rawlinson, N., de Kool, M. and Sambridge, M., 2006. Seismic wavefront tracking in 3-D heterogeneous media: applications with multiple data classes, Explor. Geophys., 37, 322-330.",
                ""
            ),
            (
                "Rawlinson, N. and Urvoy, M., 2006. Simultaneous inversion of active and passive source datasets for 3-D seismic structure with application to Tasmania, Geophys. Res. Lett., 33 L24313",
                "10.1029/2006GL028105"
            ),
            (
                "de Kool, M., Rawlinson, N. and Sambridge, M. 2006. A practical grid based method for tracking multiple refraction and reflection phases in 3D heterogeneous media, Geophys. J. Int., 167, 253-270",
                ""
            )
        ], # Reference to publication(s) that describe this example. In most 
                                # cases there will only be a single entry in this list.
                                # List of (citation, doi) pairs e.g. 
                                # citations = [("Newton, I (1687). Philosophiae naturalis principia mathematica.", None)]
                                # If there are no citations, use empty list `[]`

        "linked_sites": [("Software published on iEarth","http://iearth.edu.au/codes/FMTOMO/")],  # List of (title, address) pairs for any websites that 
                                    # should be linked in the documentation, e.g.
                                    # linked_sites = [("Parent project on Github","https://github.com/user/repo"),
                                    #                 ("Data source"),"https://www.data.com") ]
                                    # If there are no links, use empty list `[]`
    }


    def __init__(self, example_number=1):
        super().__init__(example_number)

        current_dir = path(".")
        self.tmp_files = ["fm2dss.in", "frechet.out", "gridc.vtx", "otimes.dat",
                    "raypath.out", "receivers.dat", "rtravel.out", "sources.dat",
                    "globalp.mod", "traveltime.mod"]
        self.tmp_paths = []
        for name in self.tmp_files:
            self.tmp_paths.append(current_dir / name)

        if example_number == 1:
            # read in data set
            sourcedat=np.loadtxt(path('datasets/ttimes/sources_crossb_nwt_s10.dat'))
            recdat = np.loadtxt(path('datasets/ttimes/receivers_crossb_nwt_r10.dat'))
            ttdat = np.loadtxt(path('datasets/ttimes/ttimes_crossb_nwt_s10_r10.dat'))
            recs = recdat.T[1:].T # set up receivers
            srcs = sourcedat.T[1:].T # set up sources
            nr,ns = np.shape(recs)[0],np.shape(srcs)[0] # number of receivers and sources
            print(' New data set has:\n',np.shape(recs)[0],
                ' receivers\n',np.shape(sourcedat)[0],
                ' sources\n',np.shape(ttdat)[0],' travel times')
            # rays = (ttdat[:,1] + ttdat[:,0]*nr).astype(int) # find rays from travel time file

            # Add Gaussian noise to data
            print(' Range of travel times: ',np.min(ttdat.T[2]),np.max(ttdat.T[2]),'\n Mean travel time:',np.mean(ttdat.T[2]))
            sigma =  0.00001                   # Noise is 1.0E-4 is ~5% of standard deviation of initial travel time residuals
            np.random.seed(61254557)              # set random seed
            ttdat[:,2]+=np.random.normal(0.0, sigma, len(ttdat.T[2]))

            # true model
            extent = [0.,20.,0.,30.]
            mtrue = get_gauss_model(extent,32,48) # we get the true velocity model domain for comparison 
            slowness_true = 1 / mtrue

            # starting model
            nx,ny = mtrue.shape                   # set up grid
            mb = 2000.*np.ones([nx,ny])           # reference velocity model in m/s
            slowness_starting = 1 / mb

            # assign properties
            self.exe_fm2dss = str(current_dir)
            self._mtrue = mtrue
            self._mstart = mb
            self._strue = slowness_true
            self._sstart = slowness_starting
            self._data = ttdat
            self.params["extent"] = extent
            self.params["receivers"] = recs
            self.params["sources"] = srcs
            self.params["model_shape"] = slowness_true.shape
        else:
            raise InvalidExampleError

    @property
    def description(self):
        raise NotImplementedError               # optional

    @property
    def model_size(self):
        return self.good_model.shape[0]

    @property
    def data_size(self):
        return self.data.shape[0]

    @property
    def good_model(self):
        return self._strue.flatten() 

    @property
    def starting_model(self):
        return self._sstart.flatten()
    
    @property
    def data(self):
        return self._data[:,2].flatten()

    @property
    def covariance_matrix(self):                # optional
        raise NotImplementedError

    @property
    def inverse_covariance_matrix(self):
        raise NotImplementedError               # optional

    def forward(self, model, with_jacobian=False, **kwargs): # accepting "slowness" though keyword is "model"
        slowness_reshaped = model.reshape(self._mstart.shape)
        velocity = 1 / slowness_reshaped
        g = wt.gridModel(velocity, extent=self.extent)
        if "wdir" in kwargs: kwargs.pop("wdir")
        if "frechet" in kwargs: kwargs.pop("frechet")
        fmm = g.wavefront_tracker(
            self.receivers, 
            self.sources, 
            # verbose=True, 
            # paths=True, 
            frechet=True, 
            wdir=self.exe_fm2dss,
            **kwargs,
        )
        # paths = fmm.paths
        ttimes = fmm.ttimes
        A = fmm.frechet.toarray()
        self.clean_tmp_files()
        if with_jacobian:
            return np.array(ttimes).flatten(), A
        else:
            return np.array(ttimes).flatten()
    
    def jacobian(self, model, **kwargs):      # accepting "slowness" though keyword is "model"
        return self.forward(model, True, **kwargs)[1]

    def plot_model(self, model, with_paths=False, return_paths=False, **kwargs): # accepting "slowness" though keyword is "model"
        slowness_reshaped = model.reshape(self._mstart.shape)
        velocity = 1 / slowness_reshaped
        cline = kwargs.pop("cline", "g")
        alpha = kwargs.pop("alpha", 0.5)
        if with_paths or return_paths:
            g = wt.gridModel(velocity, extent=self.extent)
            fmm = g.wavefront_tracker(
                self.receivers, 
                self.sources, 
                paths=True, 
                wdir=self.exe_fm2dss,
            )
            paths = fmm.paths
            if with_paths:
                fig = wt.displayModel(
                    velocity, 
                    paths=paths, 
                    extent=self.extent, 
                    cline=cline, 
                    alpha=alpha,
                    **kwargs
                )
            else:
                fig = wt.displayModel(
                    velocity, 
                    paths=None, 
                    extent=self.extent, 
                    cline=cline, 
                    alpha=alpha,
                    **kwargs
                )
            return (fig, paths) if return_paths else fig
        else:
            return wt.displayModel(
                velocity, 
                paths=None, 
                extent=self.extent, 
                cline=cline, 
                alpha=alpha,
                **kwargs
            ) 
    
    def plot_data(self, data, data2=None):
        raise NotImplementedError               # optional

    def misfit(self, data, data2):
        raise NotImplementedError               # optional

    def log_likelihood(self,data,data2):
        raise NotImplementedError               # optional
    
    def log_prior(self, model):
        raise NotImplementedError               # optional

    def clean_tmp_files(self):
        for file_path in self.tmp_paths:
            silent_remove(file_path)
    

def get_gauss_model(extent,nx,ny): # build two gaussian anomaly velocity model
    vc1 = 1700.                           # velocity of circle 1
    vc2 = 2300.                           # velocity of circle 2
    vb = 2000.                            # background velocity
    dx = (extent[1]-extent[0])/nx                           # cell width
    dy = (extent[3]-extent[2])/ny                           # cell height
    xc = np.linspace(extent[0],extent[1],nx)    # cell centre
    yc = np.linspace(extent[2],extent[3],ny)    # cell centre
    X,Y = np.meshgrid(xc, yc,indexing='ij')   # cell centre mesh

    # Multivariate Normal
    c1,sig1 = np.array([7.0,22.]),6.0     # location and radius of centre of first circle
    c2,sig2 = np.array([12.0,10.]),10.0    # location and radius of centre of first circle
    rv1 = multivariate_normal(c1, [[sig1, 0], [0, sig1]])
    rv2 = multivariate_normal(c2, [[sig2, 0], [0, sig2]])

    # Probability Density
    pos = np.empty(X.shape + (2,))
    pos[:, :, 0] = X
    pos[:, :, 1] = Y
    gauss1,gauss2 = rv1.pdf(pos),rv2.pdf(pos)
    return   2000.*np.ones([nx,ny])  + (vc1-vb)*gauss1/np.max(gauss1) + (vc2-vb)*gauss2/np.max(gauss2)



import numpy as np
import matplotlib.pyplot as plt

import cofi
from espresso import Magnetotelluric1D
from espresso._magnetotelluric_1D import get_frequencies, forward_1D_MT, z2rhophy, add_noise

mt = Magnetotelluric1D()

### Create synthetic data given a 5 layers Earth model

#### 0.1. Define a resistivity model and plot it
# layer electrical resitivity in log10 ohm.m 
true_model = np.array([2,1,3,1,3.5]) 
# depths in meters to the bottom of each layer (positive downwards), last layer infinite: len(true_depths)+1 = len(true_model)
true_depths = np.array([50,100,500,750])
# plot the model
max_depth = -2000
#fig = mt.plot_model(true_model, true_depths, max_depth = max_depth, title='True model')

#### 0.2. Compute response of the model
# define frequencies (in Hz) where responses are computed
fmin, fmax, f_per_decade = 1e-1, 1e4, 5
freqs = get_frequencies(fmin,fmax,f_per_decade)
# generate synthetic data 
# calculate impedance Z
Z = forward_1D_MT(true_model, true_depths, freqs, return_Z = True)
# add noise
Z, Zerr = add_noise(Z, percentage = 5, seed = 1234)
# transform Z to log10 apparent resistivity and phase (dobs)
dobs, derr = z2rhophy(freqs, Z, dZ=Zerr)
#set observed data
mt.set_obs_data(dobs, derr, freqs)
# plot the data
#fig0 = mt.plot_data(mt.data, Cm = mt.covariance_matrix)

#### 0.3. Define a starting 1D mesh and model for the inversion
# the mesh used for the inversion contains many cells to produce a smooth model
nLayers, min_thickness, vertical_growth= 80, 5, 1.1
thicknesses = [min_thickness * vertical_growth**i for i in range(nLayers)]
starting_depths = np.cumsum(thicknesses)
starting_model = np.ones((len(starting_depths)+1)) * 2 # 100 ohm.m starting model (log10 scale) 

#### 0.4. Set new starting model and mesh
mt.set_start_model(starting_model)
mt.set_start_mesh(starting_depths)


## 1. Define the problem

# define CoFI BaseProblem
mt_problem = cofi.BaseProblem()
mt_problem.set_initial_model(mt.starting_model)

# add regularization: smoothing
smoothing_factor = 50
reg_smoothing = cofi.utils.QuadraticReg(smoothing_factor, model_size = mt.model_size, reg_type = "smoothing")
reg = reg_smoothing

def objective_func(model, reg):
    dpred = mt.forward(model)
    data_misfit = mt.misfit(mt.data,dpred,mt.inverse_covariance_matrix)
    model_reg = reg(model)
    return  data_misfit + model_reg

mt_problem.set_objective(objective_func, args=[reg])


## 2. Define the inversion options

my_options = cofi.InversionOptions()
my_options.set_tool("scipy.optimize.minimize")
my_options.set_params(method="L-BFGS-B",options={'ftol':1e-3,'maxiter': 100})


## 3. Start an inversion

print("Running inversion...")
inv = cofi.Inversion(mt_problem, my_options)
inv_result = inv.run()
print("   done!")


## 4. Plotting inversion results

fig1 = mt.plot_model(inv_result.model, max_depth = max_depth, title='Smooth inversion');            # inverted model
#fig2 = mt.plot_model(true_model, true_depths, max_depth = max_depth, title='True model');       # true model
fig3 = mt.plot_data(mt.data, mt.forward(inv_result.model), Cm = mt.covariance_matrix)

nRMSE = np.sqrt(mt.misfit(mt.data, mt.forward(inv_result.model), Cm_inv = mt.inverse_covariance_matrix)/mt.data_size)
print('nRMSE = %.3f'%nRMSE)

plt.show()




########################## LIBRARY IMPORT #############################################
import io
import sys
import pathlib
import setuptools
import versioningit

try:
    import skbuild
except ImportError:
    print(
        "Please update pip, you need pip 10 or greater,\n"
        " or you need to install the PEP 518 requirements in pyproject.toml yourself",
        file=sys.stderr,
    )
    raise


_ROOT = pathlib.Path(__file__).resolve().parent

########################## VERSION ####################################################
# generate
versioningit_config = {
    "format": {
        "distance": "{base_version}+{distance}.{vcs}{rev}",
        "dirty": "{base_version}+{distance}.{vcs}{rev}.dirty",
        "distance-dirty": "{base_version}+{distance}.{vcs}{rev}.dirty",
    },
    "write": {
        "file": "src/espresso/_version.py"
    },
    "tag2version": {
        "rmprefix": "v",
        "rmsuffix": "-build",
    }
}
try:            # generate _version.py when building the core package
    versioningit.get_version(_ROOT, versioningit_config, True)
except:
    pass        # VCS not found, there should be a version file to read

# read from _version.py
with open(str(_ROOT / "src" / "espresso" / "_version.py")) as f:
    for line in f:
        if line.startswith("__version__ = "):
            _, _, version = line.partition("=")
            VERSION = version.strip(" \n'\"")
            break
    else:
        raise RuntimeError("unable to read the version from src/espresso/_version.py")


########################## LONG DESCRIPTION ###########################################
LONG_DESCRIPTION = (_ROOT / "README.md").read_text()
CONTENT_TYPE = "text/markdown"


########################## CLEAN CACHE ################################################
from shutil import rmtree
_skbuild_dir = _ROOT / "_skbuild"
if _skbuild_dir.exists() and _skbuild_dir.is_dir():
    rmtree(_skbuild_dir)


########################## OTHER METADATA #############################################
PACKAGE_NAME = "geo-espresso"
AUTHOR = f"InLab, Espresso development team"
DESCRIPTION = "Earth Science PRoblems for the Evaluation of Strategies, Solvers and Optimizers"
KEYWORDS = ["inversion", "inference", "python package", "geoscience", "geophysics"]
CLASSIFIERS = [
    "Development Status :: 2 - Pre-Alpha",
    "Intended Audience :: Education",
    "Intended Audience :: Science/Research",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: C",
    "Programming Language :: Fortran",
    "Topic :: Scientific/Engineering :: Physics",
    "License :: OSI Approved :: BSD License",
    "License :: Other/Proprietary License",
]
PACKAGE_DIR = {"": "src"}
PACKAGES = setuptools.find_namespace_packages(where='src')
CMAKE_INSTALL_DIR = f"src/espresso"
CMAKE_ARGS = ['-DSKBUILD=ON']
PYTHON_REQUIRES = ">=3.7"
INSTALL_REQUIRES = [
    "numpy>=1.18",
    "scipy>=1.0.0",
    "matplotlib>=3.1",
    "tqdm>=4.0",
    "mpmath>=1.2.0",
]


########################## SETUP ######################################################
class NullIO(io.StringIO):
    def write(self, txt):
       pass
sys.stdout = NullIO()
sys.tracebacklimit = 0

try:
    skbuild.setup(
        name=PACKAGE_NAME,
        version=VERSION,
        author=AUTHOR,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type=CONTENT_TYPE,
        keywords=KEYWORDS,
        classifiers=CLASSIFIERS,
        package_dir=PACKAGE_DIR,
        packages=PACKAGES,
        include_package_data=True,
        cmake_install_dir=CMAKE_INSTALL_DIR,
        cmake_args=CMAKE_ARGS,
        python_requires=PYTHON_REQUIRES,
        install_requires=INSTALL_REQUIRES,
    )
except SystemExit as e:
    skbuild_error : skbuild.exceptions.SKBuildError = e.args[0]
    error_message = skbuild_error.args[0]
    error_message += "\n\nHint: search 'error' in current terminal session to find out the details. "
    # error_message += f"Here are some possible reasons that cause failure in building `{PACKAGE_NAME}`:"
    # error_message += "\n\t1. no Fortran compiler found -> install a Fortran compiler and ensure it's included in the path"
    error_message += "\n\n"
    # error_message += "If the error is not due to above reasons, "
    # error_message += "please feel free to lodge an issue " \
                    #  f"at https://github.com/inlab-geo/{PACKAGE_NAME}/issues for help\n"
    skbuild_error.args = (error_message,)
    sys.exit(skbuild_error)